# -*- coding: utf-8 -*-
"""
Defines PNCursorTableModel class.
"""


from PyQt6 import QtCore, QtGui, QtWidgets  # type: ignore[import]

from pineboolib.core.utils import logging, utils_base

from sqlalchemy import exc, orm, inspect  # type: ignore [import]
from pineboolib.application.utils import date_conversion, xpm

from pineboolib.application.database.orm import utils as orm_utils
from pineboolib.application.database import pnsqlquery


import itertools
import locale
import os
import datetime


from typing import Any, Optional, List, Dict, Tuple, cast, Callable, Union, TYPE_CHECKING


if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # noqa: F401 # pragma: no cover
    from pineboolib.interfaces import (
        iconnection,
        isqlcursor,
        isqldriver,
        itablemetadata,
    )  # pragma: no cover
    from pineboolib.fllegacy import fldatatable  # pragma: no cover
    from pineboolib.application.database import pnconnectionmanager  # pragma: no cover
    from pineboolib.application.database import pnbuffer  # pragma: no cover

DEBUG = False
CURSOR_COUNT = itertools.count()
LOGGER = logging.get_logger(__name__)


class PNCursorTableModel(QtCore.QAbstractTableModel):
    """
    Link between FLSqlCursor and database.
    """

    # rows: int
    cols: int
    _use_timer = True
    where_filter: str
    where_filters: Dict[str, str] = {}
    _metadata: Optional["pntablemetadata.PNTableMetaData"]
    _sort_order = ""
    _disable_refresh: bool
    color_function_ = None
    need_update = False
    _driver_sql = None
    _show_pixmap: bool
    _size = None
    parent_view: Optional["fldatatable.FLDataTable"]
    sql_str = ""
    _can_fetch_more_rows: bool

    _parent: "isqlcursor.ISqlCursor"
    _initialized: Optional[
        bool
    ] = None  # Usa 3 estado None, True y False para hacer un primer refresh retardado si pertenece a un fldatatable
    _check_column: Dict[str, QtWidgets.QCheckBox]

    _data: List[List[Any]]
    _vdata: List[Optional[List[Any]]]

    sql_fields: List[str]
    sql_fields_omited: List[str]
    sql_fields_without_check: List[str]
    pkpos: List[int]
    ckpos: List[int]
    pkidx: Dict[Tuple, int]
    ckidx: Dict[Tuple, int]
    _column_hints: List[int]
    _current_row_data: Any
    _current_row_index: int
    _tablename: str
    _order: str

    # _data_proxy: List[Callable]
    _data_proxy: Optional["ProxyIndex"]

    _grid_obj: Dict[int, Any]

    def __init__(self, conn: "iconnection.IConnection", parent: "isqlcursor.ISqlCursor") -> None:
        """
        Initialize.

        @param conn. PNConnection Object
        @param parent. related FLSqlCursor
        """

        super(PNCursorTableModel, self).__init__()
        self._parent = parent
        self.parent_view = None

        metadata = self._parent.private_cursor.metadata_

        self.cols = 0
        self._metadata = None

        if not metadata:
            return

        self._metadata = metadata

        self._driver_sql = self._parent.db().driver()

        self.sql_fields = []
        self.sql_fields_omited = []
        self.sql_fields_without_check = []
        self.col_aliases: List[str] = []
        self._current_row_data = None
        self._current_row_index = -1

        # Indices de busqueda segun PK y CK. Los array "pos" guardan las posiciones
        # de las columnas afectadas. PK normalmente valdrá [0,].
        # CK puede ser [] o [2,3,4] por ejemplo.
        # En los IDX tendremos como clave el valor compuesto, en array, de la clave.
        # Como valor del IDX tenemos la posicion de la fila.
        # Si se hace alguna operación en _data como borrar filas intermedias hay
        # que invalidar los indices. Opcionalmente, regenerarlos.
        self.pkpos = []
        self.ckpos = []
        self.pkidx = {}
        self.ckidx = {}
        self._check_column = {}
        # Establecer a False otra vez si el contenido de los indices es erróneo.
        self.indexes_valid = False
        # self._data = []
        self._vdata = []
        self._column_hints = []
        self.updateColumnsCount()

        # self.pendingRows = 0
        # self.lastFetch = 0.0
        # self._fetched_rows = 0
        self._show_pixmap = True
        self.color_function_ = None
        # self.color_dict_ = {}

        self.where_filter = "1=1"
        self.where_filters = {}
        self.where_filters["main-filter"] = ""
        self.where_filters["filter"] = ""
        self.sql_str = ""
        self._tablename = ""
        self._order = ""

        self._disable_refresh = False
        self._initialized = None

        self._data_proxy = None

        if self.metadata().isQuery():
            query = self.conn_manager.manager().query(self.metadata().query())
            if query is None:
                raise Exception("query is empty!")
            self._tablename = query.from_()
        else:
            self._tablename = self.metadata().name()

        self._grid_obj = {}

    def disable_refresh(self, disable: bool) -> None:
        """
        Disable refresh.

        e.g. FLSqlQuery.setForwardOnly(True).
        @param disable. True or False
        """

        self._disable_refresh = disable

    def sort(
        self, column: int, order: QtCore.Qt.SortOrder = QtCore.Qt.SortOrder.AscendingOrder
    ) -> None:
        """
        Change order by used ASC/DESC and column.

        @param col. Column to sort by
        @param order. 0 ASC, 1 DESC
        """
        col = column
        # order 0 ascendente , 1 descendente
        ord = "ASC"
        if order == QtCore.Qt.SortOrder.DescendingOrder:
            ord = "DESC"

        field_mtd = self.metadata().indexFieldObject(col)
        if field_mtd.type() == "check":
            return

        col_name = field_mtd.name()

        order_list: List[str] = []
        found_ = False
        if self._sort_order:
            for column_name in self._sort_order.split(","):
                if col_name in column_name and ord in column_name:
                    found_ = True
                    order_list.append("%s %s" % (col_name, ord))
                else:
                    order_list.append(column_name)

            if not found_:
                LOGGER.debug(
                    "%s. Se intenta ordernar por una columna (%s) que no está definida en el order by previo (%s). "
                    "El order by previo se perderá" % (__name__, col_name, self._sort_order)
                )
            else:
                self._sort_order = ",".join(order_list)

        if not found_:
            self._sort_order = "%s %s" % (col_name, ord)
            self.refresh()

    def getSortOrder(self) -> str:
        """
        Get current sort order.

        Returns string  with sortOrder value.
        @return string  with info about column and order
        """
        return self._sort_order

    def setSortOrder(self, sort_order: str) -> None:
        """
        Set current ORDER BY.
        """
        # self._sort_order = ""
        self._sort_order = sort_order

    def data(
        self, index: QtCore.QModelIndex, role: int = QtCore.Qt.ItemDataRole.DisplayRole
    ) -> Any:
        """
        Retrieve information about a record.

        (overload of QAbstractTableModel)
        Could return alignment, backgroun color, value... depending on role.

        @param index. Register position
        @param role. information type required
        @return solicited data
        """

        row = index.row()
        col = index.column()
        field = self.metadata().indexFieldObject(col)
        _type = field.type()
        res_color_function: List[str] = []

        obj_ = self.get_obj_from_row(row)

        result = getattr(obj_, field.name(), None)

        if _type == "check":
            primary_key = getattr(obj_, self.metadata().primaryKey())
            if primary_key not in self._check_column.keys():
                result = QtWidgets.QCheckBox()
                self._check_column[primary_key] = result

        if self.parent_view and role in [
            QtCore.Qt.ItemDataRole.BackgroundRole,
            QtCore.Qt.ItemDataRole.ForegroundRole,
        ]:
            fun_get_color, iface = self.parent_view.functionGetColor()
            if fun_get_color is not None:
                context_ = None
                fun_name_ = None
                if fun_get_color.find(".") > -1:
                    list_ = fun_get_color.split(".")
                    from pineboolib.application.safeqsa import SafeQSA

                    qsa_widget = SafeQSA.get_any(list_[0])
                    fun_name_ = list_[1]
                    if qsa_widget:
                        context_ = qsa_widget.iface
                else:
                    context_ = iface
                    fun_name_ = fun_get_color

                function_color = getattr(context_, fun_name_, None)
                if function_color is not None:
                    field_name = field.name()
                    field_value = result
                    cursor = self._parent
                    selected = False
                    res_color_function = function_color(
                        field_name, field_value, cursor, selected, _type
                    )
                else:
                    raise Exception(
                        "No se ha resuelto functionGetColor %s desde %s" % (fun_get_color, context_)
                    )
        # print("Data ", index, role)
        # print("Registros", self.rowCount())
        # roles
        # 0 QtCore.Qt.DisplayRole
        # 1 QtCore.Qt.DecorationRole
        # 2 QtCore.Qt.EditRole
        # 3 QtCore.Qt.ToolTipRole
        # 4 QtCore.Qt.StatusTipRole
        # 5 QtCore.Qt.WhatThisRole
        # 6 QtCore.Qt.FontRole
        # 7 QtCore.Qt.TextAlignmentRole
        # 8 QtCore.Qt.BackgroundRole
        # 9 QtCore.Qt.ForegroundRole

        if role == QtCore.Qt.ItemDataRole.CheckStateRole and _type == "check":
            if primary_key in self._check_column.keys():
                if self._check_column[primary_key].isChecked():
                    return QtCore.Qt.CheckState.Checked

            return QtCore.Qt.CheckState.Unchecked

        elif role == QtCore.Qt.ItemDataRole.TextAlignmentRole:
            result = QtCore.Qt.AlignmentFlag.AlignVCenter
            if _type in ("int", "double", "uint"):
                result = result | QtCore.Qt.AlignmentFlag.AlignRight
            elif _type in ("bool", "date", "time"):
                result = result | QtCore.Qt.AlignmentFlag.AlignCenter
            elif _type in ("unlock", "pixmap"):
                result = result | QtCore.Qt.AlignmentFlag.AlignHCenter

            return result

        elif role in (QtCore.Qt.ItemDataRole.DisplayRole, QtCore.Qt.ItemDataRole.EditRole):
            if not field.visible():
                result = None
            # r = self._vdata[row]
            elif _type == "bool":
                if result in (True, "1"):
                    result = "Sí"
                else:
                    result = "No"

            elif _type in ("unlock", "pixmap"):
                result = None

            elif _type in ("string", "stringlist", "timestamp", "json"):
                if not result:
                    if _type == "stringlist":
                        result = "..."
                    else:
                        result = ""
                else:
                    result = str(result)

            elif _type == "time" and result:
                result = str(result)

            elif _type == "date":
                # Si es str lo paso a datetime.date
                if isinstance(result, str):
                    if len(result.split("-")[0]) == 4:
                        result = date_conversion.date_amd_to_dma(result)

                    if result:
                        list_ = result.split("-")
                        result = datetime.date(int(list_[2]), int(list_[1]), int(list_[0]))

                if isinstance(result, datetime.date):
                    # Cogemos el locale para presentar lo mejor posible la fecha
                    try:
                        locale.setlocale(locale.LC_TIME, "")
                        if os.name == "nt":
                            date_format = "%%d/%%m/%%y"
                        else:
                            date_format = locale.nl_langinfo(locale.D_FMT)
                        date_format = date_format.replace("y", "Y")  # Año con 4 dígitos
                        date_format = date_format.replace("/", "-")  # Separadores
                        result = result.strftime(date_format)
                    except AttributeError:
                        import platform

                        LOGGER.warning(
                            "locale specific date format is not yet implemented for %s",
                            platform.system(),
                        )

            elif _type == "check":
                return

            elif _type == "double":
                if result is not None:
                    # d = QtCore.QLocale.system().toString(float(d), "f", field.partDecimal())
                    result = utils_base.format_double(
                        result, field.partInteger(), field.partDecimal()
                    )
            elif _type in ("int", "uint"):
                if result is not None:
                    result = QtCore.QLocale.system().toString(int(result))
            if self.parent_view is not None:
                self.parent_view.resize_column(col, result)

            return result

        elif role == QtCore.Qt.ItemDataRole.DecorationRole:
            pixmap = None
            if _type in ("unlock", "pixmap") and self.parent_view:
                row_height = self.parent_view.rowHeight(row)  # Altura row
                row_width = self.parent_view.columnWidth(col)

                if _type == "unlock":
                    if result in (True, "1"):
                        pixmap = QtGui.QPixmap(
                            utils_base.filedir("./core/images/icons", "unlock.png")
                        )
                    elif result in (False, "0"):
                        pixmap = QtGui.QPixmap(
                            utils_base.filedir("./core/images/icons", "lock.png")
                        )
                else:
                    if not self._parent.private_cursor._is_system_table:
                        data = self.conn_manager.manager().fetchLargeValue(result)
                    else:
                        data = xpm.cache_xpm(result)  # type: ignore [arg-type]

                    pixmap = QtGui.QPixmap(data)
                    if not pixmap.isNull():
                        new_size = row_height - 1
                        if new_size > row_width:
                            new_size = row_width

                        pixmap = pixmap.scaled(new_size, new_size)

                if self.parent_view.showAllPixmap() or row == self.parent_view.cur.at():
                    if pixmap and not pixmap.isNull() and self.parent_view:
                        new_pixmap = QtGui.QPixmap(row_width, row_height)  # w , h
                        center_width = (row_width - pixmap.width()) / 2
                        center_height = (row_height - pixmap.height()) / 2
                        new_pixmap.fill(QtCore.Qt.GlobalColor.transparent)
                        painter = QtGui.QPainter(new_pixmap)
                        painter.drawPixmap(
                            int(center_width),
                            int(center_height),
                            pixmap.width(),
                            pixmap.height(),
                            pixmap,
                        )

                        pixmap = new_pixmap

            return pixmap

        elif role == QtCore.Qt.ItemDataRole.BackgroundRole:
            if _type == "bool":
                if result in (True, "1"):
                    result = QtGui.QBrush(QtCore.Qt.GlobalColor.green)
                else:
                    result = QtGui.QBrush(QtCore.Qt.GlobalColor.red)

            elif _type == "check":
                obj_ = self._check_column[primary_key]  # type: ignore [assignment]
                result = (
                    QtGui.QBrush(QtCore.Qt.GlobalColor.green)
                    if obj_.isChecked()  # type: ignore [union-attr]
                    else QtGui.QBrush(QtCore.Qt.GlobalColor.white)
                )

            else:
                if res_color_function and len(res_color_function) and res_color_function[0] != "":
                    color_ = QtGui.QColor(res_color_function[0])
                    style_ = getattr(QtCore.Qt, res_color_function[2])
                    result = QtGui.QBrush(color_)
                    result.setStyle(style_)
                else:
                    result = None

            return result

        elif role == QtCore.Qt.ItemDataRole.ForegroundRole:
            if _type == "bool":
                if result in (True, "1"):
                    result = QtGui.QBrush(QtCore.Qt.GlobalColor.black)
                else:
                    result = QtGui.QBrush(QtCore.Qt.GlobalColor.white)
            else:
                if res_color_function and len(res_color_function) and res_color_function[1] != "":
                    color_ = QtGui.QColor(res_color_function[1])
                    style_ = getattr(QtCore.Qt, res_color_function[2])
                    result = QtGui.QBrush(color_)
                    result.setStyle(style_)
                else:
                    result = None

            return result

        return None

    def update_rows(self) -> None:
        """Update virtual records managed by its model."""

        parent = QtCore.QModelIndex()
        to_row = self.rowCount() - 1

        self.beginInsertRows(parent, 0, to_row)
        self.endInsertRows()
        top_left = self.index(0, 0)
        botom_rigth = self.index(to_row, self.cols - 1)
        self.dataChanged.emit(top_left, botom_rigth)  # type: ignore [attr-defined] # noqa: F821
        self.indexes_valid = True

    def _refresh_field_info(self) -> None:
        """
        Check if query fields do exist.

        If any field does not exist it gets marked as ommitted.
        """
        is_query = self.metadata().isQuery()
        qry_tables: List[
            Tuple[
                str,
                Optional[Union["itablemetadata.ITableMetaData", "pntablemetadata.PNTableMetaData"]],
            ]
        ] = []
        qry = None

        if is_query:
            qry_file = self.metadata().query()
            qry = self.conn_manager.manager().query(qry_file)
            if qry is None:
                LOGGER.error(
                    "Could not load the file %s.qry for an unknown reason. This table is a view",
                    qry_file,
                )
                raise Exception(
                    "Could not load the file %s.qry for an unknown reason. This table is a view"
                    % qry_file
                )
            qry_select = [x.strip() for x in (qry.select()).split(",")]
            qry_fields: Dict[str, str] = {
                fieldname.split(".")[-1]: fieldname for fieldname in qry_select
            }

            # for table in qry.tablesList():
            #    mtd = self.conn_manager.manager().metadata(table, True)
            #    if mtd:
            #        qry_tables.append((table, mtd))

            qry_tables = [
                (table, self.conn_manager.manager().metadata(table, True))
                for table in qry.tablesList()
            ]

        for number, field in enumerate(self.metadata().fieldList()):
            # if field.visibleGrid():
            #    sql_fields.append(field.name())
            if field.isPrimaryKey():
                self.pkpos.append(number)
            if field.isCompoundKey():
                self.ckpos.append(number)

            if is_query:
                if field.name() in qry_fields:
                    self.sql_fields.append(qry_fields[field.name()])
                else:
                    found = False
                    for table_name, table_meta in qry_tables:
                        if table_meta and field.name() in table_meta.fieldNames():
                            self.sql_fields.append("%s.%s" % (table_name, field.name()))
                            found = True
                            break
                    # Omito los campos que aparentemente no existen
                    if not found and not field.name() in self.sql_fields_omited:
                        if qry is None:
                            raise Exception("The qry is empty!")

                        # NOTE: Esto podría ser por ejemplo porque no entendemos los campos computados.
                        LOGGER.error(
                            "CursorTableModel.refresh(): Omitiendo campo '%s' referenciado en query %s. El campo no existe en %s ",
                            field.name(),
                            self.metadata().name(),
                            qry.tablesList(),
                        )
                        self.sql_fields_omited.append(field.name())

            else:
                if field.type() != field.Check:
                    self.sql_fields_without_check.append(field.name())

                self.sql_fields.append(field.name())

    def insert_current_buffer(self) -> bool:
        """Insert data from current buffer."""
        try:
            obj_ = self.buffer.current_object()
            self.session.add(obj_)
            orm_utils.do_flush(self.session, [obj_])
            return True
        except Exception as error:
            LOGGER.warning("insert_current_buffer : %s" % error, stack_info=True)

        return False

    # def edit_current_buffer(self):
    #    """Update data from current buffer."""

    def delete_current_buffer(self) -> bool:
        """Delete data from current buffer."""

        try:
            obj_ = self.buffer.current_object()
            self.session.delete(obj_)
            orm_utils.do_flush(self.session, [obj_])
            return True
        except Exception as error:
            LOGGER.warning("delete_current_buffer : %s" % error, stack_info=True)

        return False

    def refresh(self) -> None:
        """
        Refresh information mananged by this class.
        """

        if utils_base.is_library():
            self._initialized = False

        if (
            self._initialized is None and self.parent_view
        ):  # Si es el primer refresh y estoy conectado a un FLDatatable()
            self._initialized = True
            QtCore.QTimer.singleShot(1, self.refresh)
            return

        if (
            self._initialized
        ):  # Si estoy inicializando y no me ha enviado un sender, cancelo el refesh
            if not self.sender():
                return

        self._initialized = False
        rows = self.rowCount()

        if (self._disable_refresh or self._parent.modeAccess() == self._parent.Del) and rows:
            return

        self._grid_obj = {}
        self._parent.clear_buffer()

        where_filter = self.buildWhere()

        # """ FIN """

        parent = QtCore.QModelIndex()

        self.beginRemoveRows(parent, 0, rows)
        self.endRemoveRows()
        if rows > 0:
            cast(
                QtCore.pyqtSignal, self.rowsRemoved
            ).emit(  # type: ignore [attr-defined] # noqa: F821
                parent, 0, rows - 1
            )

        self._refresh_field_info()

        # if self.metadata().isQuery():
        #    print("FIXME!! query!!")

        # dynamic_filter_class = sql_tools.DynamicFilter(
        #    query=session_.query(self._parent._cursor_model), model_class=self._parent._cursor_model
        # )

        # dynamic_filter_class.set_filter_condition_from_string(where_filter)

        # self._data_proxy = dynamic_filter_class.return_query()
        if self.metadata().isQuery():
            meta_qry = pnsqlquery.PNSqlQuery(self.metadata().query())
            if where_filter.strip().lower().startswith("order"):
                order_by = where_filter.lower().replace("order by", "")
                if order_by:
                    meta_qry.setOrderBy(order_by)
                where_filter = "1 = 1"

            meta_qry.setWhere(where_filter.replace("WHERE ", ""))
            sql_query = meta_qry.sql()
        else:
            sql_query = "SELECT %s FROM %s %s" % (
                self.metadata().primaryKey(),
                self.metadata().name(),
                where_filter,
            )

        self._data_proxy = None
        # print("COUNT", sql_count)

        # print("QUERY", sql_query)
        result_query = self.session.execute(sql_query)  # type: ignore [arg-type]
        rows_loaded = result_query.rowcount  # type: ignore [attr-defined]

        if rows_loaded == -1:  # Si rowcount no está disponible...
            sql_count = (
                "SELECT COUNT(%s) FROM " % self.metadata().primaryKey()
                + sql_query[sql_query.find(" FROM ") + 6 :]
            )

            if sql_count.find("ORDER BY") > 1:
                if sql_count.find("WHERE") == -1:
                    sql_count += " WHERE 1 = 1"
                sql_count = sql_count[: sql_count.find("ORDER BY")]

            result_count = self.session.execute(sql_count)  # type: ignore [arg-type]
            rows_loaded = result_count.fetchone()[0]  # type: ignore [index]

        if rows_loaded > 0:
            self._data_proxy = ProxyIndex(result_query, rows_loaded)

            self.need_update = False
            self._column_hints = [120] * len(self.sql_fields)

        if self.parent_view:
            self.update_rows()

    def buildWhere(self) -> str:
        """Return valid where."""

        where_filter = " AND ".join(
            [value.strip() for key, value in sorted(self.where_filters.items()) if value]
        )

        if self._sort_order:
            if where_filter.find("ORDER BY") == -1:
                order_by = " ORDER BY %s" % self._sort_order
                if where_filter.find(";") > -1:
                    where_filter = where_filter.replace(";", "%s;" % order_by)
                else:
                    where_filter += order_by

        return (
            where_filter
            if where_filter.strip().startswith("ORDER")
            else "WHERE %s" % where_filter
            if where_filter
            else ""
        )

    def updateCacheData(self, mode: int) -> bool:
        """Update cache data without refresh."""
        # print("* updateCacheData", mode)
        # mode 1- Insert, 2 - Edit, 3 - Del

        if self._disable_refresh:
            return True

        pk_name = self._parent.primaryKey()
        pk_value = self._parent.buffer().value(pk_name)
        where_filter = self.buildWhere()

        sql_query = "SELECT %s FROM %s %s" % (
            self.metadata().primaryKey(),
            self.metadata().name(),
            where_filter,
        )
        order_by = ""
        if sql_query.find("ORDER BY") > 1:
            sql_query = sql_query[: sql_query.find("ORDER BY")]
            order_by = sql_query[sql_query.find("ORDER BY") :]

        extra_where = " %s" % (
            self.conn_manager.manager().formatAssignValue(self.metadata().field(pk_name), pk_value)
        )
        if extra_where not in sql_query:
            sql_query += " AND" if sql_query.find("WHERE") > -1 else " WHERE"
            sql_query += extra_where

        result = self.session.execute(sql_query)  # type: ignore [arg-type]
        new_data = result.fetchone()

        if new_data is None and mode in [1, 2]:  # mode 3 allways returns None
            LOGGER.debug("no valid data to update cache!")
            return True

        if self._data_proxy is None:
            LOGGER.debug("data_proxy is empty!")
            return True
        elif (
            mode == 1 and new_data[0] in self._data_proxy._cached_data  # type: ignore [index]
        ):  # if exists dont need Insert.
            return True

        if mode == 1:  # Insert.
            if order_by:
                LOGGER.warning("FIXME! update chache whit alternative order_by")
                return False
            else:
                current_pos = None
                min_val = 0
                max_val = self._data_proxy._total_rows

                while True:
                    upper = None

                    if self.rowCount():
                        if current_pos is None:
                            current_pos = max_val // 2

                        # while current_pos > self._data_proxy._last_current_size:
                        #    if not self._data_proxy.fetch_more():
                        #        break

                        # if current_pos < self._data_proxy._last_current_size:
                        #    data = self._data_proxy[current_pos]
                        data = self._data_proxy[current_pos]

                        if self._data_proxy._last_current_size < current_pos:
                            LOGGER.warning(
                                "Error seek possition %s over %s (len %s). Total: %s"
                                % (
                                    current_pos,
                                    self._data_proxy._qry_rows_loaded,
                                    self._data_proxy._last_current_size,
                                    self._data_proxy._total_rows,
                                )
                            )
                            return False

                        if pk_value > data:
                            if current_pos in (max_val, 0):
                                upper = True
                            else:
                                min_val = current_pos
                                current_pos += (max_val - min_val) // 2
                        else:
                            if current_pos in (min_val, 0):
                                upper = False
                            else:
                                max_val = current_pos
                                current_pos -= (max_val - min_val) // 2

                        if (max_val - min_val) // 2 == 0:
                            upper = True
                    elif self._data_proxy is not None:
                        upper = False
                        current_pos = 0

                    if upper is not None:
                        if upper:
                            current_pos += 1

                        self._data_proxy.insert(
                            new_data,
                            current_pos
                            if current_pos < self._data_proxy._last_current_size
                            else -1,
                        )

                        break
            return True

        elif mode == 2:  # Edit.
            self._data_proxy.update_pk(pk_value, new_data[0])  # type: ignore [index]

            return True

        elif mode == 3:  # Delete.
            self._data_proxy.delete_pk(pk_value)

            return True

        LOGGER.warning("%s.model.updateCacheData(%s) returns False" % (self._parent._name, mode))
        return False

    def get_obj_from_row(self, row: int) -> Optional[Callable]:
        """Return row object from proxy."""

        if row > -1 and row < self.rowCount() and self._data_proxy:
            pk_value = self._data_proxy[row]
            session_ = self.session

            if (
                row not in self._grid_obj.keys()
                or self._grid_obj[row] is not None
                and inspect(self._grid_obj[row]).expired
            ):
                query = orm_utils.DynamicFilter(
                    query=session_.query(self._parent._cursor_model),
                    model_class=self._parent._cursor_model,
                )
                query.set_filter_condition_from_string(
                    "%s = %s"
                    % (self.metadata().primaryKey(), str(pk_value).replace(" ", "_|_space_|_"))
                )
                try:
                    self._grid_obj[row] = query.return_query().first()
                except Exception as error:
                    raise Exception(
                        "get_object_from_row %s (%s) : %s" % (row, pk_value, error)
                    ) from error

            return self._grid_obj[row]

        return None

    def seek_row(self, row: int) -> bool:
        """Seek row selected."""

        if not hasattr(self, "_current_row_index") or row != self._current_row_index:
            if row > -1 and row < self.rowCount():
                object_ = self.get_obj_from_row(row)
                if object_ is None:
                    return False
                self._current_row_index = row
                self._current_row_data = object_

            else:
                return False

        return True

    def value(self, row: int, field_name: str) -> Any:
        """Return colum value from a row."""

        return (
            getattr(self.get_obj_from_row(row), field_name, None)
            if row > -1 and row < self.rowCount()
            else None
        )

    def find_pk_row(self, pk_value: Any) -> int:
        """Retrieve row index of a record given a primary key."""

        try:
            return self._data_proxy.index(pk_value) if self._data_proxy else -1
        except ValueError:
            return -1

    def fieldType(self, field_name: str) -> str:
        """
        Retrieve field type for a given field name.

        @param field_name. required field name.
        @return field type.
        """
        field = self.metadata().field(field_name)
        if field is None:
            raise Exception("field %s not found" % field_name)
        return field.type()

    def alias(self, field_name: str) -> str:
        """
        Retrieve alias name for a field name.

        @param field_name. field name requested.
        @return alias for the field.
        """
        field = self.metadata().field(field_name)
        if field is None:
            raise Exception("field %s not found" % field_name)
        return field.alias()

    def columnCount(self, *args: List[Any]) -> int:  # type: ignore [override] # noqa F821
        """
        Get current column count.

        @return Number of columns present.
        """
        # if args:
        #    LOGGER.warning("columnCount%r: wrong arg count", args, stack_info=True)
        return self.cols

    def updateColumnsCount(self) -> None:
        """
        Set number of columns in tableModel.
        """
        self.cols = len(self.metadata().fieldList())
        self.loadColAliases()
        if self.metadata().isQuery():
            self._refresh_field_info()

    def rowCount(self, parent: Optional["QtCore.QModelIndex"] = None) -> int:
        """
        Get current row count.

        @return Row number present in table.
        """
        return (
            self._data_proxy._total_rows  # type: ignore [union-attr] # noqa: F821
            if getattr(self, "_data_proxy", None)
            else 0
        )

    def headerData(
        self,
        section: int,
        orientation: "QtCore.Qt.Orientation",
        role: int = QtCore.Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        """
        Retrieve header data.

        @param section. Column
        @param orientation. Horizontal, Vertical
        @param role. QtCore.Qt.ItemDataRole.DisplayRole only. Every other option is ommitted.
        @return info for section, orientation and role.
        """
        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                if not self.col_aliases:
                    self.loadColAliases()
                return self.col_aliases[section]
            elif orientation == QtCore.Qt.Orientation.Vertical:
                return section + 1
        return None

    def loadColAliases(self) -> None:
        """
        Load column alias for every column.
        """
        self.col_aliases = [
            str(self.metadata().indexFieldObject(i).alias()) for i in range(self.cols)
        ]

    def metadata(self) -> "pntablemetadata.PNTableMetaData":
        """
        Retrieve FLTableMetaData for this tableModel.

        @return Objeto FLTableMetaData
        """
        if self._metadata is None:
            raise Exception("metadata is empty!")

        return self._metadata

    @property
    def db(self) -> "iconnection.IConnection":
        """Get current connection."""

        return self._parent.db()

    @property
    def buffer(self) -> "pnbuffer.PNBuffer":
        """Get buffer."""

        return self._parent.buffer()

    @property
    def driver_sql(self) -> "isqldriver.ISqlDriver":
        """Return driver sql."""

        return self.db.driver()

    @property
    def session(self) -> "orm.Session":
        """Return pnsqlcursor session."""

        return self.db.session()

    @property
    def conn_manager(self) -> "pnconnectionmanager.PNConnectionManager":
        """Return connection manager."""

        return self.db.connManager()

    def set_parent_view(self, parent_view: "fldatatable.FLDataTable") -> None:
        """Set the parent view."""
        self.parent_view = parent_view


class ProxyIndex:
    """ProxyIndex class."""

    _query = None
    _cached_data: List[Any] = []
    _index: int
    _total_rows: int
    _qry_rows_total: int
    _qry_rows_loaded: int
    _last_current_size: int

    def __init__(self, result_query: Any, rows: int) -> None:
        """Initialize."""

        self._query = result_query
        self._qry_rows_total = self._total_rows = int(rows)

        self._qry_rows_loaded = 0
        self._last_current_size = 0
        self._cached_data = []

        self.fetch_more(2000 if rows > 2000 else rows)

    def __getitem__(self, index: int) -> Any:
        """Return item value."""

        if self._last_current_size <= index:
            self.fetch_more(index - self._last_current_size + 1)

        return self._cached_data[index] if self._last_current_size > index else None

    def index(self, value: Any) -> int:
        """Return data position."""

        while value not in self._cached_data:
            if not self.fetch_more():
                return -1

        return self._cached_data.index(value)

    def fetch_more(self, fetch_size: int = 2000) -> bool:
        """Fetch more data to cached data."""

        if self._qry_rows_loaded < self._qry_rows_total and self._query:
            to_fetch = self._qry_rows_loaded + fetch_size
            fetch_size = (
                (self._qry_rows_total - self._qry_rows_loaded)
                if to_fetch > 0 and to_fetch >= self._qry_rows_total
                else fetch_size
            )

            try:
                self._cached_data += [data[0] for data in self._query.fetchmany(fetch_size)]
                self._qry_rows_loaded += fetch_size
                self._last_current_size += fetch_size
                return True
            except exc.InterfaceError:
                LOGGER.warning(
                    "Se ha producido un problema al recoger %s primary keys del caché. cacheadas: %s, totales: %s",
                    fetch_size,
                    self._qry_rows_loaded,
                    self._qry_rows_total,
                )

        return False

    def insert(self, value: Any, position: int = -1) -> bool:
        """Insert a value to cache."""

        if position == -1:
            self._cached_data.append(value)
        else:
            self._cached_data.insert(position, value)

        self._last_current_size += 1
        self._total_rows += 1

        return True

    def update_pk(self, pk_value: Any, value: Any) -> bool:
        """Update a cached value."""

        self._cached_data[self.index(pk_value)] = value

        return True

    def delete_pk(self, pk_value: Any) -> bool:
        """Delete a cached value."""

        index = self.index(pk_value)
        if index > -1:
            del self._cached_data[index]
            self._total_rows -= 1
            self._last_current_size -= 1

            return True

        return False



"""Qtexstream module."""

from PyQt6 import QtCore  # type: ignore[import]


class QTextStream(QtCore.QTextStream):
    """QTextStream class."""

    def opIn(self, text_):
        """Set value to QTextStream."""
        self.device().write(text_.encode())  # type: ignore [union-attr]

    def read(self, max_len: int = 0) -> str:
        """Read datas from QTextStream."""

        if max_len > 0:
            return super().read(max_len)
        else:
            return super().readAll()



"""
Module for MYISAM driver.
"""
from pineboolib import logging
from pineboolib.interfaces import isqldriver
from typing import Any, Optional, List, Dict, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.application.metadata import pntablemetadata  # noqa: F401 # pragma: no cover
    from sqlalchemy.engine import base  # type: ignore [import] # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FLMYSQL_MYISAM(isqldriver.ISqlDriver):  # pylint: disable=invalid-name
    """MYISAM Driver class."""

    _default_charset: str

    def __init__(self):
        """Create empty driver."""
        super().__init__()
        self.version_ = "0.9"
        self.name_ = "FLMYSQL_MyISAM"
        self.alias_ = "MySQL MyISAM (MYSQLDB)"
        self._no_inno_db = True
        self.default_port = 3306
        self.active_create_index = True
        self._true = True
        self._false = False
        self._like_true = "1"
        self._like_false = "0"
        self._text_like = " "
        self._create_isolation = False
        self._use_create_table_save_points = False
        self._database_not_found_keywords = ["Unknown database"]
        self._default_charset = "DEFAULT CHARACTER SET = utf8 COLLATE = utf8_bin"
        self._sqlalchemy_name = "mysql"

    def tables(self, type_name: str = "", table_name: str = "") -> List[str]:
        """Return a tables list specified by type."""
        table_list: List[str] = []
        result_list: List[Any] = []
        if self.is_open():
            where: List[str] = []
            if type_name in ("Tables", ""):
                where.append(
                    "TABLE_TYPE LIKE 'BASE TABLE' AND TABLE_SCHEMA LIKE '%s'" % self.DBName()
                )
            if type_name in ("Views", ""):
                where.append("TABLE_TYPE LIKE 'VIEW' AND TABLE_SCHEMA LIKE '%s'" % self.DBName())
            if type_name in ("SystemTables", ""):
                where.append("TABLE_TYPE LIKE 'SYSTEM VIEW'")

            if where:
                and_name = " AND TABLE_NAME ='%s'" % (table_name) if table_name else ""

                cursor = self.execute_query(
                    "SELECT TABLE_NAME,TABLE_TYPE,ENGINE FROM information_schema.tables where %s%s ORDER BY TABLE_NAME ASC"
                    % (" OR ".join(where), and_name)
                )
                result_list += cursor.fetchall() if cursor else []

            for item in result_list:
                table_name = item[0]
                table_list.append(table_name)
                if table_name not in self._tables_info.keys():
                    self._tables_info[table_name] = {
                        "type": item[1],
                        "engine": item[2],
                    }

            # table_list = [item[0] for item in result_list]

        return table_list

    def setType(self, type_: str, leng: int = 0) -> str:
        """Return type definition."""
        res_ = ""
        type_ = type_.lower()
        if type_ == "int":
            res_ = "INTEGER"
        elif type_ in ["uint", "serial"]:
            res_ = "INT UNSIGNED"
        elif type_ in ("bool", "unlock"):
            res_ = "BOOL"
        elif type_ == "double":
            res_ = "DECIMAL"
        elif type_ == "time":
            res_ = "TIME"
        elif type_ == "date":
            res_ = "DATE"
        elif type_ in ["pixmap", "stringlist"]:
            res_ = "MEDIUMTEXT"
        elif type_ == "string":
            if leng > 255:
                res_ = "VARCHAR"
            else:
                res_ = "CHAR"
        elif type_ == "bytearray":
            res_ = "LONGBLOB"
        elif type_ == "timestamp":
            res_ = "TIMESTAMP"
        elif type_ == "json":
            res_ = "JSON"
        else:
            LOGGER.warning("seType: unknown type %s", type_)
            leng = 0

        return "%s(%s)" % (res_, leng) if leng else res_

    def sqlCreateTable(
        self,
        tmd: "pntablemetadata.PNTableMetaData",
        create_index: bool = True,
        is_view: bool = False,
    ) -> Optional[str]:
        """Create a table from given MTD."""

        if tmd.isQuery():
            return self.sqlCreateView(tmd)

        from pineboolib.fllegacy import flutil

        util = flutil.FLUtil()
        primary_key = ""
        sql = "CREATE TABLE %s (" % tmd.name()
        # seq = None

        field_list = tmd.fieldList()

        unlocks = 0
        for number, field in enumerate(field_list):
            type_ = field.type()
            leng_ = field.length()
            if field.type() == "unlock":
                unlocks += 1

                if unlocks > 1:
                    LOGGER.debug("FLManager : No se ha podido crear la tabla " + tmd.name())
                    LOGGER.debug(
                        "FLManager : Hay mas de un campo tipo unlock. Solo puede haber uno."
                    )
                    return None

            sql += field.name()
            if type_ == "double":
                sql += " DECIMAL(%s,%s)" % (
                    field.partInteger() + field.partDecimal() + 5,
                    field.partDecimal() + 5,
                )

            else:
                if type_ == "string" and leng_ == 0:
                    leng_ = 255

                sql += " %s" % self.setType(type_, leng_)

            if field.isPrimaryKey():
                if not primary_key:
                    sql += " PRIMARY KEY"
                    primary_key = field.name()
                else:
                    LOGGER.warning(
                        util.translate(
                            "application",
                            "FLManager : Tabla-> %s ." % tmd.name()
                            + "Se ha intentado poner una segunda clave primaria para el campo %s ,pero el campo %s ya es clave primaria."
                            % (primary_key, field.name())
                            + "Sólo puede existir una clave primaria en FLTableMetaData, use FLCompoundKey para crear claves compuestas.",
                        )
                    )
                    raise Exception(
                        "A primary key (%s) has been defined before the field %s.%s -> %s"
                        % (primary_key, tmd.name(), field.name(), sql)
                    )
            else:
                sql += " UNIQUE" if field.isUnique() else ""
                sql += " NULL" if field.allowNull() else " NOT NULL"

            if number != len(field_list) - 1:
                sql += ","

        engine = ") ENGINE=INNODB" if not self._no_inno_db else ") ENGINE=MyISAM"
        sql += engine

        sql += " %s" % self._default_charset

        return sql

    def dict_cursor(self) -> Any:
        """Return dict cursor."""

        from MySQLdb.cursors import DictCursor  # type: ignore

        return DictCursor

    def recordInfo2(self, table_name: str) -> Dict[str, List[Any]]:
        """Obtain current cursor information on columns."""

        info = {}

        sql = "SHOW FIELDS FROM %s" % table_name

        cursor = self.execute_query(sql)
        res = cursor.fetchall() if cursor else []

        for columns in res:
            field_name = columns[0]
            field_allow_null = columns[2] == "NO"
            field_size = "0"
            field_precision = 0
            field_default_value = columns[4]
            field_primary_key = columns[3] == "PRI"

            if columns[1].find("(") > -1:
                field_type = self.decodeSqlType(columns[1][: columns[1].find("(")])
                if field_type not in ["uint", "int", "double"]:
                    field_size = columns[1][columns[1].find("(") + 1 : columns[1].find(")")]
                else:
                    pos_comma = field_size.find(",")
                    if pos_comma > -1:
                        list_number = field_size.split(",")
                        field_precision = int(list_number[1])
                        field_size = list_number[0]

            else:
                field_type = self.decodeSqlType(columns[1])

            if field_type == "string" and field_size == "255":
                field_size = "0"

            info[field_name] = [
                field_name,
                field_type,
                field_allow_null,
                int(field_size),
                field_precision,
                field_default_value,
                field_primary_key,
            ]

        return info

    def decodeSqlType(self, t: str) -> str:
        """Translate types."""
        ret = t

        if t in ["char", "varchar", "text"]:
            ret = "string"
        elif t == "int unsigned":
            ret = "uint"
        elif t == "int":
            ret = "int"
        elif t == "date":
            ret = "date"
        elif t in ["mediumtext"]:
            ret = "stringlist"
        elif t == "tinyint":
            ret = "bool"
        elif t in ["decimal", "double"]:
            ret = "double"
        elif t == "longblob":
            ret = "bytearray"
        elif t == "time":
            ret = "time"
        elif t == "timestamp":
            ret = "timestamp"
        elif t in ["json", "longtext"]:
            ret = "json"

        else:
            LOGGER.warning("formato desconocido %s", ret)

        return ret

    def vacuum(self):
        """Vacuum tables."""
        table_names = self.db_.tables("Tables")
        # self._connection.connection.set_isolation_level(0)
        for table_name in table_names:
            if self.db_.connManager().manager().metadata(table_name) is not None:
                self.execute_query("ANALYZE TABLE %s" % table_name)
        # self._connection.connection.set_isolation_level(1)

    def getAlternativeConn(
        self, name: str, host: str, port: int, usern: str, passw_: str
    ) -> Optional["base.Connection"]:
        """Return alternative connection."""
        return self.getConn("", host, port, usern, passw_)

    def do_connect(self, dbapi_connection, connection_record):
        """Isolation Level fix."""

        dbapi_connection.isolation_level = "READ COMMITTED"

    def get_common_params(self) -> None:
        """Load common params."""

        super().get_common_params()
        self._queqe_params["isolation_level"] = "READ COMMITTED"

        if self.pool_enabled():
            self._queqe_params["pool_recycle"] = 1800
        else:
            if "connect_args" not in self._queqe_params.keys():
                self._queqe_params["connect_args"] = {}
            if "connect_timeout" not in self._queqe_params["connect_args"].keys():
                self._queqe_params["connect_args"]["connect_timeout"] = 28800  # 8 horas.

    def connect(
        self, db_name: str, db_host: str, db_port: int, db_user_name: str, db_password: str
    ) -> Union["base.Connection", bool]:
        """Connect to database."""

        if self._no_inno_db:
            LOGGER.warning(
                "MyISAM engines doesn't support transactions. ROLLBACK statement doesn't produce any effect"
            )

        return super().connect(db_name, db_host, db_port, db_user_name, db_password)

    def mismatchedTable(self, table_name: str, metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """Return if a table is mismatched."""

        if self.invalid_engine(table_name):
            return True

        return super().mismatchedTable(table_name, metadata)

    def invalid_engine(self, table_name: str, mute: bool = False) -> bool:
        """Return if table engine is valid."""

        engine_name = self._tables_info[table_name]["engine"]
        if not self._no_inno_db == (engine_name == "MyISAM"):
            if not mute:
                LOGGER.warning(
                    "The engine of the %s table is of type %s, but the driver uses the %s engine",
                    table_name,
                    engine_name,
                    "MyISAM" if self._no_inno_db else "InnoDB",
                )
            return True

        return False



"""Test_pnsqlquery module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.application.database import pnsqlquery, pnsqlcursor
from pineboolib import application
from pineboolib.application.database.tests import fixture_path


class TestPNSqlQuery1(unittest.TestCase):
    """TestPNSqlQuery1 Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test basic_1."""
        from pineboolib.application.database import pnparameterquery, pngroupbyquery

        qry = pnsqlquery.PNSqlQuery("fltest2")

        from_param = pnparameterquery.PNParameterQuery("from", "from", 2)
        to_param = pnparameterquery.PNParameterQuery("to", "to", 2)
        from_param.setValue(0)
        to_param.setValue(1)
        qry.addParameter(from_param)
        qry.addParameter(to_param)
        data = {}
        data[from_param.name()] = from_param.value()
        data[to_param.name()] = to_param.value()
        qry.setParameterDict(data)

        self.assertEqual(qry.groupDict(), {0: "fltest.id"})

        self.assertEqual(qry.valueParam("to"), 1)
        qry.setValueParam("to", 2)
        self.assertEqual(qry.valueParam("to"), 2)
        self.assertEqual(
            qry.sql(),
            "SELECT id,string_field,date_field,time_field,double_field,bool_field,uint_field,bloqueo,empty_relation,int_field FROM fltest"
            + " WHERE id>='0' AND id<='2' ORDER BY fltest.id",
        )

        gr_01 = pngroupbyquery.PNGroupByQuery(0, "string_field")

        qry.addGroup(gr_01)
        group = {}
        group[gr_01.level()] = gr_01.field()

        qry2 = pnsqlquery.PNSqlQuery("fltest")
        qry2.setSelect(
            "SUM(id),id,string_field,date_field,time_field,double_field,bool_field,uint_field,bloqueo"
        )
        qry2.setFrom("fltest")
        qry2.setWhere("id>='0' AND id<='1'")
        qry2.setGroupDict(group)

        self.assertEqual(
            qry2.sql(),
            "SELECT SUM(id),id,string_field,date_field,time_field,double_field,bool_field,uint_field,bloqueo FROM fltest"
            + " WHERE id>='0' AND id<='1' ORDER BY string_field",
        )

        self.assertEqual(qry.name(), "fltest2")
        self.assertEqual(qry.where(), "id>=[from] AND id<=[to]")
        self.assertEqual(qry.orderBy(), "fltest.id")
        qry.setSelect(["SUM(id)", "id", "fltest.string_field"])
        self.assertEqual(len(qry.parameterDict()), 2)
        self.assertEqual(len(qry.groupDict()), 1)
        self.assertEqual(qry.fieldList(), ["SUM(id)", "id", "fltest.string_field"])
        self.assertEqual(qry.posToFieldName(0), "SUM(id)")
        self.assertEqual(qry.posToFieldName(1), "id")
        self.assertEqual(qry.posToFieldName(2), "fltest.string_field")
        self.assertEqual(qry.fieldNameToPos("fltest.string_field"), 2)
        self.assertEqual(qry.fieldNameToPos("string_field"), 2)
        qry.setName("fltest2_dos")

        self.assertEqual(qry.name(), "fltest2_dos")
        self.assertEqual(len(qry.fieldMetaDataList()), 3)

    def test_basic_2(self) -> None:
        """Test basic_2."""

        qry = pnsqlquery.PNSqlQuery("fake")
        qry.setTablesList("fake_table")
        self.assertEqual(qry.tablesList(), ["fake_table"])
        qry.setTablesList(["fake_table_1", "fake_table_2"])
        self.assertEqual(qry.tablesList(), ["fake_table_1", "fake_table_2"])
        qry.setSelect("field_01")
        qry.setFrom("fake_table")
        qry.setWhere("1=1")
        qry.setOrderBy("field_01 ASC")
        self.assertEqual(
            qry.sql(), "SELECT field_01 FROM fake_table WHERE 1=1 ORDER BY field_01 ASC"
        )

        self.assertEqual(qry.fieldNameToPos("field_01"), 0)

        self.assertFalse(qry.exec_())
        self.assertEqual(qry.fieldList(), ["field_01"])
        self.assertFalse(qry.isValid())
        self.assertTrue(qry.isNull("field_01"))
        self.assertEqual(qry.value("field_01"), None)
        qry.showDebug()

    def test_basic_3(self) -> None:
        """Test basic_3."""

        qry = pnsqlquery.PNSqlQuery("fake")
        qry.setTablesList("fake_table")
        qry.setSelect("field_01")
        qry.setFrom("fake_table")
        qry.setWhere("1=1")
        qry.setOrderBy("field_01 ASC")
        self.assertEqual(
            qry.sql(), "SELECT field_01 FROM fake_table WHERE 1=1 ORDER BY field_01 ASC"
        )
        self.assertFalse(qry.isForwardOnly())
        qry.setForwardOnly(True)
        self.assertTrue(qry.isForwardOnly())
        qry.setForwardOnly(False)
        self.assertFalse(qry.isForwardOnly())
        self.assertFalse(qry.lastError())
        qry2 = pnsqlquery.PNSqlQuery("fake")
        self.assertFalse(qry2.exec_("SELEFT * FROM DDD"))
        self.assertTrue(qry.lastError())
        self.assertEqual(qry2.driver(), qry2.db().driver())
        self.assertEqual(qry2.numRowsAffected(), 0)
        self.assertTrue(qry2.lastQuery(), "SELEFT * FROM DDD")
        self.assertFalse(qry2.isValid())
        self.assertFalse(qry2.isActive())

    def test_case(self) -> None:
        """Test case."""

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("idarea", "G")
        cursor.setValueBuffer("descripcion", "Area G")
        cursor.commitBuffer()

        qry = pnsqlquery.PNSqlQuery("flareas")
        qry.setSelect("idarea, case when idarea ='G' THEN 'YO' ELSE 'TU' END")
        qry.setFrom("flareas")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(
            qry.fieldList(), ["idarea", "case when idarea ='g' then 'yo' else 'tu' end"]
        )
        self.assertEqual(qry.value("case when idarea ='G' THEN 'YO' ELSE 'TU' END"), "YO")

        qry2 = pnsqlquery.PNSqlQuery("flareas")
        qry2.setSelect("idarea, case when idarea ='G' THEN 'YO' ELSE ((((5 * 4) + 2))/2) END")
        qry2.setFrom("flareas")
        qry2.setWhere("1=1")
        self.assertTrue(qry2.exec_())
        self.assertTrue(qry2.first())
        self.assertEqual(
            qry2.fieldList(),
            ["idarea", "case when idarea ='g' then 'yo' else ((((5 * 4) + 2))/2) end"],
        )
        self.assertEqual(
            qry2.value("case when idarea ='G' THEN 'YO' ELSE ((((5 * 4) + 2))/2) END"), "YO"
        )
        self.assertFalse(
            qry2.isNull("case when idarea ='g' then 'yo' else ((((5 * 4) + 2))/2) end")
        )

    def test_move(self) -> None:
        """Test move functions."""

        cursor_6 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_6.setModeAccess(cursor_6.Insert)
        cursor_6.refreshBuffer()
        cursor_6.setValueBuffer("bloqueo", True)
        cursor_6.setValueBuffer("idarea", "O")
        cursor_6.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor_6.commitBuffer())
        cursor_6.setModeAccess(cursor_6.Insert)
        cursor_6.refreshBuffer()
        cursor_6.setValueBuffer("bloqueo", True)
        cursor_6.setValueBuffer("idarea", "P")
        cursor_6.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor_6.commitBuffer())
        cursor_6.commit()

        qry = pnsqlquery.PNSqlQuery("")
        qry.setTablesList("flareas")
        qry.setSelect("idarea")
        qry.setFrom("flareas")
        qry.setWhere("1=1")
        qry.setOrderBy("idarea ASC")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        val_first = qry.value(0)
        size_ = qry.size()
        self.assertTrue(qry.last())
        val_last = qry.value(0)
        self.assertNotEqual(qry.value("idarea"), val_first)
        self.assertEqual(qry.value("idarea"), qry.value(0))
        self.assertTrue(qry.prev())
        self.assertTrue(qry.seek(0))
        self.assertFalse(qry.isNull("idarea"))
        self.assertEqual(qry.value(0), val_first)
        self.assertFalse(qry.seek(1000))
        self.assertFalse(qry.seek(1000, True))
        self.assertTrue(qry.seek(size_ - 1, True))  # last
        self.assertEqual(qry.value(0), val_last)

    def test_parentesis(self) -> None:
        """Test parentesis."""

        qry = pnsqlquery.PNSqlQuery("fake")
        qry.exec_(
            "select CAST(MAX(codagente) as INTEGER), campo_dos from agentes where codagente ~ '^[0-9]+$' AND codagente < 99990"
        )
        self.assertEqual(qry.tablesList(), ["agentes"])
        self.assertEqual(qry.from_(), "agentes")
        self.assertEqual(qry.fieldList(), ["cast (max(codagente) as integer)", "campo_dos"])

    def test_sql_injection(self) -> None:
        """Test sql injection."""

        qry = pnsqlquery.PNSqlQuery("fake")

        qry.sql_inspector._check_sql_injection(["email", "=", "'ncastanaresrodriguez@gmail.com'"])
        self.assertFalse(
            qry.sql_inspector.suspected_injection(),
            "SOSPECHOSO : %s" % qry.sql_inspector._suspected_injection,
        )

    def test_only_inspector(self) -> None:
        """Test only inspector."""

        qry = pnsqlquery.PNSqlQuery("fake")
        qry.exec_(
            "SELECT SUM(munitos), dia, noche FROM dias WHERE astro = 'sol' GROUP BY dias.minutos ORDER BY dia ASC, noche DESC"
        )
        self.assertEqual(qry.tablesList(), ["dias"])
        self.assertEqual(qry.from_(), "dias")
        self.assertEqual(qry.fieldList(), ["sum(munitos)", "dia", "noche"])
        self.assertEqual(qry.select(), "sum(munitos),dia,noche")
        self.assertEqual(qry.orderBy(), "dia asc, noche desc")
        self.assertEqual(qry.where(), "astro = 'sol'")

        qry_2 = pnsqlquery.PNSqlQuery("fake")
        qry_2.exec_(
            "SELECT SUM(munitos), dia, noche, p.nombre FROM dias INNER JOIN planetas AS "
            + "p ON p.id = dias.id WHERE astro = 'sol' GROUP BY dias.minutos ORDER BY dia ASC, noche DESC"
        )
        self.assertEqual(qry_2.tablesList(), ["dias", "planetas"])
        self.assertEqual(qry_2.fieldNameToPos("planetas.nombre"), 3)
        self.assertEqual(qry_2.fieldList(), ["sum(munitos)", "dia", "noche", "p.nombre"])
        self.assertEqual(qry_2.fieldNameToPos("nombre"), 3)
        self.assertEqual(qry_2.fieldNameToPos("p.nombre"), 3)
        self.assertEqual(qry_2.posToFieldName(3), "p.nombre")
        self.assertEqual(qry_2.posToFieldName(2), "noche")
        self.assertEqual(qry_2.where(), "astro = 'sol'")
        self.assertEqual(qry_2.from_(), "dias inner join planetas as p on p.id = dias.id")

    def test_date_result(self) -> None:
        """Test date values."""
        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("date_field", "2020-01-01")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.commit()

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("date_field")
        qry.setFrom("fltest")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.next())
        self.assertTrue(qry.isNull("date_field"))
        self.assertEqual(qry.value(0), "")
        self.assertEqual(qry.value("date_field"), "")
        self.assertTrue(qry.next())
        self.assertFalse(qry.isNull("date_field"))
        self.assertEqual(str(qry.value(0)), "2020-01-01T00:00:00")
        self.assertEqual(str(qry.value("date_field")), "2020-01-01T00:00:00")
        self.assertTrue(qry.next())
        self.assertTrue(qry.isNull("date_field"))
        self.assertEqual(qry.value(0), "")
        self.assertEqual(qry.value("date_field"), "")

    def test_limit_offset(self) -> None:
        """Test limit and offset clausules from a query."""

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())  # 9 rows total!
        cursor.commit()

        qry_one = pnsqlquery.PNSqlQuery()
        qry_one.setSelect("date_field")
        qry_one.setFrom("fltest")
        qry_one.setWhere("1 = 1")
        qry_one.setLimit(4)
        self.assertTrue(qry_one.exec_())
        self.assertTrue(qry_one.sql().lower().find("limit") > -1)
        self.assertEqual(qry_one.size(), 4)

        qry_two = pnsqlquery.PNSqlQuery()
        qry_two.setSelect("date_field")
        qry_two.setFrom("fltest")
        qry_two.setWhere("1 = 1")
        qry_two.setLimit(100)
        qry_two.setOffset(7)
        self.assertTrue(qry_two.exec_())
        self.assertTrue(qry_two.sql().lower().find("offset") > -1)
        self.assertEqual(qry_two.size(), 2)  # 7 + 2 = 9 rows

        qry_tree = pnsqlquery.PNSqlQuery()
        qry_tree.setSelect("date_field")
        qry_tree.setFrom("fltest")
        qry_tree.setWhere("1 = 1")
        qry_tree.setOrderBy("date_field")
        qry_tree.setOffset(5)
        self.assertTrue(qry_tree.exec_())
        sql = qry_tree.sql()
        self.assertTrue(sql.lower().find("offset") > -1)
        self.assertTrue(sql.lower().find("order by") > -1)
        self.assertEqual(qry_tree.size(), 4)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestPNSqlQuery2(unittest.TestCase):
    """TestPNSqlQuery2 Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        application.LOG_SQL = False
        init_testing()

    def test_basic_4(self) -> None:
        """Test basic test 4."""
        from pineboolib.qsa import qsa
        from pineboolib.application.metadata import pntablemetadata, pnfieldmetadata
        import os

        cur_date_str = str(qsa.Date())
        qsa_sys = qsa.sys

        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))

        widget = qsa.from_project("flfactppal")
        widget.iface.valoresIniciales()
        cur_clientes = qsa.FLSqlCursor("clientes")
        cur_clientes.setModeAccess(cur_clientes.Insert)
        cur_clientes.refreshBuffer()
        # cur_clientes.setValueBuffer("codigo", "000001")
        cur_clientes.setValueBuffer("nombre", "cliente de prueba")
        cur_clientes.setValueBuffer("cifnif", "01234567H")
        cur_clientes.setValueBuffer("codserie", "A")
        self.assertTrue(cur_clientes.commitBuffer())
        cur_clientes.commit()
        # cur_clientes.conn().doTransaction(cur_clientes)
        mtd_tareas = pntablemetadata.PNTableMetaData("tareas")
        field_01 = pnfieldmetadata.PNFieldMetaData(
            "idtarea",
            "Id",
            False,
            True,
            "serial",
            0,
            True,
            True,
            True,
            0,
            0,
            True,
            True,
            False,
            None,
            False,
            False,
            False,
            True,
            False,
        )
        field_02 = pnfieldmetadata.PNFieldMetaData(
            "nombre",
            "Nombre",
            False,
            False,
            "string",
            10,
            False,
            True,
            True,
            0,
            0,
            False,
            False,
            False,
            None,
            False,
            False,
            True,
            False,
            False,
        )
        field_03 = pnfieldmetadata.PNFieldMetaData(
            "fechaini",
            "Fecha Inicial",
            True,
            False,
            "date",
            0,
            False,
            True,
            True,
            0,
            0,
            False,
            False,
            False,
            None,
            False,
            False,
            True,
            False,
            False,
        )
        field_04 = pnfieldmetadata.PNFieldMetaData(
            "fechafinal",
            "Fecha Final",
            True,
            False,
            "date",
            0,
            False,
            True,
            True,
            0,
            0,
            False,
            False,
            False,
            None,
            False,
            False,
            True,
            False,
            False,
        )
        mtd_tareas.addFieldMD(field_01)
        mtd_tareas.addFieldMD(field_02)
        mtd_tareas.addFieldMD(field_03)
        mtd_tareas.addFieldMD(field_04)
        self.assertEqual(
            mtd_tareas.fieldListArray(False), ["idtarea", "nombre", "fechaini", "fechafinal"]
        )
        application.PROJECT.conn_manager.manager().cache_metadata_["tareas.mtd"] = mtd_tareas
        self.assertTrue(application.PROJECT.conn_manager.manager().createTable("tareas"))
        self.assertTrue(application.PROJECT.conn_manager.manager().existsTable("tareas"))

        cur_tareas = qsa.FLSqlCursor("tareas")
        self.assertEqual(
            cur_tareas.metadata().fieldListArray(False),
            ["idtarea", "nombre", "fechaini", "fechafinal"],
        )

        cur_tareas.setModeAccess(cur_tareas.Insert)
        cur_tareas.refreshBuffer()
        # cur_tareas.setValueBuffer("idtarea", 1)
        cur_tareas.setValueBuffer("nombre", "prueba1")
        self.assertTrue(cur_tareas.commitBuffer())
        cur_tareas.setModeAccess(cur_tareas.Insert)
        cur_tareas.refreshBuffer()
        # cur_tareas.setValueBuffer("idtarea", 2)
        cur_tareas.setValueBuffer("nombre", "prueba2")
        cur_tareas.setValueBuffer("fechaini", cur_date_str)
        cur_tareas.setValueBuffer("fechafinal", cur_date_str)
        self.assertTrue(cur_tareas.commitBuffer())
        cur_tareas.setModeAccess(cur_tareas.Insert)
        cur_tareas.refreshBuffer()
        # cur_tareas.setValueBuffer("idtarea", 3)
        cur_tareas.setValueBuffer("nombre", "prueba3")
        cur_tareas.setValueBuffer("fechaini", cur_date_str)
        cur_tareas.setValueBuffer("fechafinal", cur_date_str)
        self.assertTrue(cur_tareas.commitBuffer())
        cur_tareas.setModeAccess(cur_tareas.Insert)
        cur_tareas.refreshBuffer()
        cur_tareas.setValueBuffer("nombre", "prueba4")
        cur_tareas.setValueBuffer("fechaini", cur_date_str)
        self.assertTrue(cur_tareas.commitBuffer())
        cur_tareas.setModeAccess(cur_tareas.Insert)
        cur_tareas.refreshBuffer()
        # cur_tareas.setValueBuffer("idtarea", 3)
        cur_tareas.setValueBuffer("nombre", "prueba5")
        cur_tareas.setValueBuffer("fechafinal", cur_date_str)
        self.assertTrue(cur_tareas.commitBuffer())
        cur_tareas.commit()

        qry = qsa.FLSqlQuery()
        qry.setSelect("idtarea,nombre,fechaini,fechafinal")
        qry.setFrom("tareas")
        qry.setWhere("1=1")
        qry.setOrderBy("idtarea")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        self.assertEqual(qry.value("fechaini"), "")
        self.assertEqual(qry.value("fechafinal"), "")
        self.assertTrue(qry.next())
        self.assertNotEqual(qry.value("fechaini"), "")
        self.assertNotEqual(qry.value("fechafinal"), "")
        self.assertTrue(qry.next())
        self.assertNotEqual(qry.value("fechaini"), "")
        self.assertNotEqual(qry.value("fechafinal"), "")
        self.assertTrue(qry.next())
        self.assertNotEqual(qry.value("fechaini"), "")
        self.assertEqual(qry.value("fechafinal"), "")
        self.assertTrue(qry.next())
        self.assertEqual(qry.value("fechaini"), "")
        self.assertNotEqual(qry.value("fechafinal"), "")

    def test_basic_5(self) -> None:
        """Test query without where."""

        from pineboolib.qsa import qsa

        qry = qsa.FLSqlQuery()
        qry.setSelect("agentes.codagente,agentes.nombre")
        qry.setFrom("agentes")
        qry.setOrderBy("agentes.codagente ASC LIMIT 11")
        self.assertTrue(qry.exec_())

    def test_invalid_tables(self) -> None:
        """Test invalid tables."""

        qry = pnsqlquery.PNSqlQuery()
        qry.setSelect("area.idarea,modulo.idmodelo")
        qry.setFrom(
            "flareas area\n\tINNER JOIN flmodules modulo ON (area.idarea = CAST(modulo.idarea AS STRING) AND modulo.bloqueado = False"
        )
        qry.setWhere("NOT modulo.bloqueado")

        qry.sql_inspector.set_sql(qry.sql())
        qry.sql_inspector.resolve()
        self.assertFalse(qry.sql_inspector._invalid_tables)
        self.assertTrue(qry.isValid())

        qry.exec_("select 1")
        self.assertTrue(qry.isValid())

        qry2 = pnsqlquery.PNSqlQuery()
        qry2.setSelect("area.idarea,modulo.idmodelo")
        qry2.setFrom(
            """flareas area\n\tINNER JOIN flmodules modulo ON (area.idarea = CAST (modulo.idarea AS STRING)
            AND CAST(modulo.bloqueado AS BOOL) = False"""
        )
        qry2.setWhere("NOT modulo.bloqueado")

        qry2.sql_inspector.set_sql(qry2.sql())
        qry2.sql_inspector.resolve()
        self.assertFalse(qry2.sql_inspector._invalid_tables)
        self.assertTrue(qry2.isValid())

        # qry2.exec_("select 1")
        # self.assertTrue(qry2.isValid())

    def test_sql_injection(self) -> None:
        sql = (
            "SELECT s.codalmacen,t.descripcion,t.direccion,t.ciudad,t.provincia,t.codpostal,t.codpais,"
            + "t.telefono, s.talla FROM tpv_tiendas t inner join stocks s on t.codalmacen = s.codalmacen "
            + "left outer join param_parametros p on 'RSTOCK_' || t.codalmacen = p.nombre WHERE "
            + "t.sincroactiva and (p.nombre like 'RSTOCK_%' or p.nombre is null) AND s.disponible > 1 AND "
            + "s.referencia = '4070W200050' AND t.idempresa not in (15,42,44) AND s.talla = '-1'"
            + " OR ASCII(SUBSTRING((SELECT/**/COALESCE(CAST(current_database()/**/AS/**/CHARACTER(10000))"
            + ",(CHR(32))))::text/**/FROM/**/43/**/FOR/**/1))>119 AND 000622=000622 or 'yFfSTINR'='' GROUP"
            + " BY s.codalmacen,t.descripcion,t.direccion,t.ciudad,t.provincia,t.codpostal,t.codpais,t.telefono,"
            + " s.talla ORDER BY s.codalmacen, s.talla"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        self.assertTrue(qry.sql_inspector.suspected_injection())

    def test_or_and_not_in_where(self) -> None:
        """Test or and not in where."""

        sql = (
            "SELECT crm_contactos.codcontacto,crm_contactos.nombre,crm_contactos.email,crm_contactos.telefono1,"
            + "crm_contactos.codagente,COUNT(ss_tratos.idtrato),SUM(ss_tratos.valor) FROM crm_contactos "
            + "LEFT OUTER JOIN ss_tratos ON crm_contactos.codcontacto = ss_tratos.codcontacto AND "
            + "crm_contactos.codagente = ss_tratos.codagente AND (ss_tratos.estado is null or ss_tratos.estado "
            + "NOT IN ('Ganado','Perdido')) WHERE ((crm_contactos.nombre ILIKE '%%%%' OR crm_contactos.email ILIKE '%%%%'"
            + " OR crm_contactos.telefono1 ILIKE '%%%%') AND crm_contactos.codagente IN ('555')) "
            + "GROUP BY crm_contactos.codcontacto,crm_contactos.nombre,crm_contactos.email,crm_contactos.telefono1,"
            + "crm_contactos.codagente ORDER BY crm_contactos.codcontacto ASC LIMIT 51"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        self.assertEqual(
            qry.sql_inspector.field_list(),
            {
                "crm_contactos.codcontacto": 0,
                "crm_contactos.nombre": 1,
                "crm_contactos.email": 2,
                "crm_contactos.telefono1": 3,
                "crm_contactos.codagente": 4,
                "count(ss_tratos.idtrato)": 5,
                "sum(ss_tratos.valor)": 6,
            },
        )
        self.assertEqual(qry.sql_inspector.table_names(), ["crm_contactos", "ss_tratos"])

    def test_interval(self) -> None:
        """Test interval special word."""

        sql = (
            "SELECT fam.codfamilia, fam.descripcion, SUM(lf.pvptotal) as facturacion"
            + " FROM familias fam INNER JOIN articulos a ON fam.codfamilia = a.codfamilia"
            + " INNER JOIN lineasfacturascli lf ON lf.referencia = a.referencia INNER JOIN"
            + " facturascli f ON lf.idfactura = f.idfactura AND f.fecha > "
            + "(current_date - INTERVAL '12 months') GROUP BY fam.codfamilia, fam.descripcion"
            + " HAVING SUM(lf.pvptotal) > 0 ORDER BY facturacion DESC;"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        print("*", qry.sql_inspector.table_names())
        self.assertEqual(
            qry.sql_inspector.table_names(),
            ["familias", "articulos", "lineasfacturascli", "facturascli"],
        )

    def test_as_in_select(self) -> None:
        """Test as in select."""
        sql = (
            "SELECT idpedido, "
            + "SUM(CASE WHEN operacion = 'C' THEN importe ELSE 0 END) AS cobros, "
            + "SUM(CASE WHEN operacion = 'V' THEN importe ELSE  0 END) AS ventas "
            + " FROM coe_cobrosventasped "
            + " WHERE idpedido = "
            + "str(idpedido)"
            + " GROUP BY idpedido"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        self.assertEqual(qry.sql_inspector.field_list(), {"idpedido": 0, "cobros": 1, "ventas": 2})

        sql = (
            "SELECT idpedido, "
            + "SUM(CASE WHEN operacion = 'C' THEN importe ELSE 0 END) AS cobros, "
            + "SUM(CASE WHEN operacion = 'V' THEN importe ELSE ( 0 ) END) AS ventas "
            + " FROM coe_cobrosventasped "
            + " WHERE idpedido = "
            + "str(idpedido)"
            + " GROUP BY idpedido"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        self.assertEqual(qry.sql_inspector.field_list(), {"idpedido": 0, "cobros": 1, "ventas": 2})

    def test_distinct_on(self) -> None:
        """Test distinct in select."""
        sql = (
            "select distinct on (l.referencia) l.referencia,p.fecha,(l.cantidad - l.totalenalbaran)"
            + " from pedidosprov p inner join lineaspedidosprov l on p.idpedido = l.idpedido where"
            + " p.codproveedor = '000802' and p.codalmacen = 'alm' and l.referencia in "
            + "('00608004','00608005','00609100','00609300','00614003','00614004','00622001'"
            + ",'00622003','00622107','00627900','01809001') and p.servido <> 'sí' and"
            + " l.cantidad > l.totalenalbaran order by l.referencia, p.fecha"
        )

        qry = pnsqlquery.PNSqlQuery()
        qry.sql_inspector.set_sql(sql)
        qry.sql_inspector.resolve()
        self.assertEqual(
            qry.sql_inspector.field_list(),
            {"l.referencia": 0, "p.fecha": 1, "(l.cantidad - l.totalenalbaran)": 2},
        )

    def test_resolve_fields(self) -> None:
        """Test resolve fields."""

        qry = pnsqlquery.PNSqlQuery()
        por_dto = 10
        qry.setSelect("SUM((pvptotal * iva * (100 - %s)) / 100 / 100), iva" % (por_dto))
        qry.setFrom("lineasalbaranescli")
        qry.setWhere("1=1")

        qry.sql_inspector.set_sql(qry.sql())
        qry.sql_inspector.resolve()
        self.assertEqual(
            qry.sql_inspector.field_list(),
            {"sum((pvptotal * iva * (100 - 10)) / 100 / 100)": 0, "iva": 1},
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()
        application.LOG_SQL = False



"""
Tests for Class on qsa.
"""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestClass(unittest.TestCase):
    """Test Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test basic."""

        class_ = qsa.class_.Prueba
        self.assertEqual(qsa.class_.classes(), ["Prueba"])
        self.assertTrue(class_)
        self.assertEqual(class_.__name__, "Prueba")
        self.assertNotEqual(class_(), class_())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""Fllineedit module."""

from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore[import]
from pineboolib import logging
from typing import Optional

LOGGER = logging.get_logger(__name__)


class FLLineEdit(QtWidgets.QLineEdit):
    """FLLineEdit class."""

    _tipo: str
    _part_decimal: int
    _part_integer: int
    _max_value: int
    _auto_select: bool
    _name: str
    _longitud_max: int
    _parent: QtWidgets.QWidget
    _last_text: str
    _formating: bool
    _field_name: str

    lostFocus = QtCore.pyqtSignal()

    def __init__(self, parent: QtWidgets.QWidget, name: str = "") -> None:
        """Inicialize."""
        super().__init__(parent)
        self._name = name
        field_name = getattr(parent, "_field_name", None)
        if field_name is not None:
            self._field_name = field_name
            cursor = getattr(parent, "cursor_", None)
            if cursor is not None:
                mtd = cursor.metadata()
                if mtd is None:
                    raise Exception("mtd is Empty!")

                self._tipo = mtd.field(self._field_name).type()
                self._part_decimal = 0
                self._auto_select = True
                self._formating = False
                self._part_integer = mtd.field(self._field_name).partInteger()

                self._parent = parent

                if self._tipo == "string":
                    self._longitud_max = mtd.field(self._field_name).length()
                    self.setMaxLength(self._longitud_max)

                elif self._tipo in ("int", "uint", "double"):
                    self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)

    def setText(self, text_: str, check_focus: bool = True) -> None:  # type: ignore [override]
        """Set text to control."""

        text_ = str(text_)
        # if not project.DGI.localDesktop():
        #    project.DGI._par.addQueque("%s_setText" % self._parent.objectName(), text_)
        # else:
        if check_focus:
            if text_ in ("", None) or self.hasFocus():
                super().setText(text_)
                return

        ok_ = False

        minus = False

        if self._tipo == "double":
            if text_[0] == "-":
                minus = True
                text_ = text_[1:]

            val, ok_ = QtCore.QLocale.system().toDouble(text_.replace(".", ","))  # type: ignore[assignment]

            if ok_:
                text_ = QtCore.QLocale.system().toString(float(text_), "f", self._part_decimal)
            if minus:
                text_ = "-%s" % text_

        elif self._tipo in ("int"):
            val, ok_ = QtCore.QLocale.system().toInt(text_)  # type: ignore[assignment]
            if ok_:
                text_ = QtCore.QLocale.system().toString(val)

        elif self._tipo in ("uint"):
            val, ok_ = QtCore.QLocale.system().toUInt(text_)  # type: ignore[assignment]
            if ok_:
                text_ = QtCore.QLocale.system().toString(val)

        super().setText(str(text_))

    def text(self) -> str:
        """Return text from control."""

        text_ = super().text()
        if text_ == "":
            return text_

        ok_ = False
        minus = False

        if self._tipo == "double":
            if text_[0] == "-":
                minus = True
                text_ = text_[1:]

            val, ok_ = QtCore.QLocale.system().toDouble(text_)  # type: ignore[assignment]
            if ok_:
                text_ = str(val)

            if minus:
                text_ = "-%s" % text_

            if text_ == ",":
                text_ = ""

        elif self._tipo == "uint":
            val, ok_ = QtCore.QLocale.system().toUInt(text_)  # type: ignore[assignment]
            if ok_:
                text_ = str(val)

        elif self._tipo == "int":
            val, ok_ = QtCore.QLocale.system().toInt(text_)  # type: ignore[assignment]

            if ok_:
                text_ = str(val)

        return text_

    def setMaxValue(self, max_value: int) -> None:
        """Set max value for numeric types."""

        self._max_value = max_value

    def focusOutEvent(self, event: Optional["QtGui.QFocusEvent"]) -> None:
        """Focus out event."""

        if self._tipo in ("double", "int", "uint"):
            text_ = super().text()

            if self._tipo == "double":
                val, ok_ = QtCore.QLocale.system().toDouble(text_)

                if ok_:
                    text_ = QtCore.QLocale.system().toString(val, "f", self._part_decimal)
                super().setText(text_)
            else:
                self.setText(text_)
        super().focusOutEvent(event)  # type: ignore [arg-type]

    def focusInEvent(self, event: Optional["QtGui.QFocusEvent"]) -> None:
        """Focus in event."""

        if self.isReadOnly():
            return

        if self._tipo in ("double", "int", "uint"):
            self.blockSignals(True)
            s_orig = self.text()
            text_ = s_orig
            if self._tipo == "double":
                if s_orig != "":
                    text_ = QtCore.QLocale.system().toString(float(s_orig), "f", self._part_decimal)

                if QtCore.QLocale.system().toString(1.1, "f", 1)[1] == ",":
                    text_ = text_.replace(".", "")
                else:
                    text_ = text_.replace(",", "")

            validator = self.validator()
            if validator:
                pos = 0
                validator.validate(text_, pos)

            super().setText(text_)
            self.blockSignals(False)

        if self._auto_select and not self.selectedText() and not self.isReadOnly():
            self.selectAll()

        super().focusInEvent(event)  # type: ignore [arg-type]



"""Test_qradiobuttons module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing

from pineboolib.q3widgets import qhbuttongroup
from pineboolib.q3widgets import qvbuttongroup
from pineboolib.q3widgets import qradiobutton


class TestQRadioButtons(unittest.TestCase):
    """Test TestQRadioButtons class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Test basic 1."""

        hbg = qhbuttongroup.QHButtonGroup()
        vbg = qvbuttongroup.QVButtonGroup()

        self.assertTrue(hbg)
        self.assertTrue(vbg)

        bt1 = qradiobutton.QRadioButton(hbg)
        bt1.text = "prueba"
        bt1.setButtonGroupId(10)
        self.assertFalse(bt1.checked)
        bt1.checked = True
        bt1.send_clicked()
        self.assertTrue(bt1.checked)
        self.assertEqual(bt1.text, "prueba")
        hbg.setSelectedId(10)
        self.assertEqual(hbg.selectedId, 10)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
AQSql Module.

Use the resources of pineboolib.application.database.pnsqlquery.PNSqlQuery.
"""

from pineboolib.application.database.pnsqlquery import PNSqlQuery


class AQSqlQuery(PNSqlQuery):
    """AQSql Class."""

    pass



"""
Tests for signatures package.
"""

import os.path


def fixture_path(*path: str) -> str:
    """
    Get fixture path for this test folder.
    """
    basedir = os.path.realpath(os.path.dirname(__file__))
    filepath = os.path.join(basedir, "fixtures", *path)
    return filepath



"""Flmodules module."""
# -*- coding: utf-8 -*-
from pineboolib import logging
from pineboolib.qsa import qsa

from pineboolib.application.parsers.parser_qsa import postparse
import os
import pathlib
from typing import Union, TYPE_CHECKING


if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # type: ignore[import] # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def init(self) -> None:
        """Init function."""
        btn_load = self.child("botonCargar")
        btn_export = self.child("botonExportar")
        self.module_connect(btn_load, "clicked()", self, "load_button_clicked")
        self.module_connect(btn_export, "clicked()", self, "expoort_button_clicked")
        cursor = self.cursor()
        if cursor.modeAccess() == cursor.Browse:
            btn_load.setEnabled(False)
            btn_export.setEnabled(False)

    def load_file_to_db(
        self,
        nombre: str,
        contenido: Union[str, bytes],
        log: "QtWidgets.QTextEdit",
        directorio: str,
    ) -> None:
        """Load a file into database."""
        if not qsa.util.isFLDefFile(contenido) and not nombre.endswith(  # type: ignore [arg-type]
            (
                ".mod",
                ".xpm",
                ".signatures",
                ".checksum",
                ".certificates",
                ".qs",
                ".ar",
                ".py",
                ".kut",
                ".jasper",
                ".jrxml",
            )
        ):
            return

        cursor_ficheros = qsa.FLSqlCursor("flfiles")
        cursor = self.cursor()

        binary = nombre.endswith(".jasper")
        cursor_ficheros.select(qsa.ustr("nombre = '", nombre, "'"))
        if not cursor_ficheros.first():
            if nombre.endswith(".ar"):
                if not self.load_ar(nombre, contenido, log, directorio):  # type: ignore [arg-type]
                    return
            log.append(qsa.util.translate("scripts", "- Cargando :: ") + nombre)
            cursor_ficheros.setModeAccess(cursor_ficheros.Insert)
            cursor_ficheros.refreshBuffer()
            cursor_ficheros.setValueBuffer("nombre", nombre)
            cursor_ficheros.setValueBuffer("idmodulo", cursor.valueBuffer("idmodulo"))
            if binary:
                cursor_ficheros.setValueBuffer(
                    "sha", qsa.util.sha1(contenido.decode(errors="replace"))  # type: ignore [union-attr]
                )
                cursor_ficheros.setValueBuffer("binario", contenido)
            else:
                cursor_ficheros.setValueBuffer("sha", qsa.util.sha1(contenido))
                cursor_ficheros.setValueBuffer("contenido", contenido)
            cursor_ficheros.commitBuffer()

        else:
            cursor_ficheros.setModeAccess(cursor_ficheros.Edit)
            cursor_ficheros.refreshBuffer()
            copy_content = cursor_ficheros.valueBuffer("contenido")
            copy_binario = cursor_ficheros.valueBuffer("binario")
            if (not binary and copy_content != contenido) or (
                binary
                and copy_binario.decode(errors="replace") != contenido.decode(errors="replace")  # type: ignore [union-attr]
            ):
                log.append(qsa.util.translate("scripts", "- Actualizando :: ") + nombre)
                cursor_ficheros.setModeAccess(cursor_ficheros.Insert)
                cursor_ficheros.refreshBuffer()
                this_date = qsa.Date()
                cursor_ficheros.setValueBuffer("nombre", nombre + qsa.parseString(this_date))
                cursor_ficheros.setValueBuffer("idmodulo", cursor.valueBuffer("idmodulo"))
                cursor_ficheros.setValueBuffer("contenido", copy_content)
                cursor_ficheros.setValueBuffer("binario", copy_binario)
                cursor_ficheros.commitBuffer()
                log.append(
                    qsa.util.translate("scripts", "- Backup :: ")
                    + nombre
                    + qsa.parseString(this_date)
                )
                cursor_ficheros.select(qsa.ustr("nombre = '", nombre, "'"))
                cursor_ficheros.first()
                cursor_ficheros.setModeAccess(cursor_ficheros.Edit)
                cursor_ficheros.refreshBuffer()
                cursor_ficheros.setValueBuffer("idmodulo", cursor.valueBuffer("idmodulo"))
                if binary:
                    cursor_ficheros.setValueBuffer(
                        "sha", qsa.util.sha1(contenido.decode(errors="replace"))  # type: ignore [union-attr]
                    )
                    cursor_ficheros.setValueBuffer("binario", contenido)
                else:
                    cursor_ficheros.setValueBuffer("sha", qsa.util.sha1(contenido))
                    cursor_ficheros.setValueBuffer("contenido", contenido)
                cursor_ficheros.commitBuffer()
                if nombre.endswith(".ar"):
                    self.load_ar(nombre, contenido, log, directorio)  # type: ignore [arg-type]

        # cursor_ficheros.close()

    def load_ar(
        self, nombre: str, contenido: str, log: "QtWidgets.QTextEdit", directorio: str
    ) -> bool:
        """Load AR reports."""
        if not qsa.sys.isLoadedModule("flar2kut"):
            return False
        if qsa.util.readSettingEntry("scripts/sys/conversionAr") != "true":
            return False
        log.append(qsa.util.translate("scripts", "Convirtiendo %s a kut") % (str(nombre)))
        contenido = qsa.sys.toUnicode(contenido, "UTF-8")
        contenido = qsa.from_project("flar2kut").iface.pub_ar2kut(contenido)
        nombre = qsa.ustr(qsa.parseString(nombre)[0 : len(nombre) - 3], ".kut")
        if contenido:
            local_encode = qsa.util.readSettingEntry("scripts/sys/conversionArENC", "ISO-8859-15")
            contenido = qsa.sys.fromUnicode(contenido, local_encode)
            self.load_file_to_db(nombre, contenido, log, directorio)
            log.append(qsa.util.translate("scripts", "Volcando a disco ") + nombre)
            qsa.FileStatic.write(qsa.Dir.cleanDirPath(qsa.ustr(directorio, "/", nombre)), contenido)

        else:
            log.append(qsa.util.translate("scripts", "Error de conversión"))
            return False

        return True

    def load_files(self, directorio: str, extension: str) -> None:
        """Load files into database."""

        log = self.child("log")
        if log is None:
            raise Exception("log is empty!.")

        extension = extension.replace("*", "")

        settings = qsa.FLSettings()
        parse_modules_on_load = settings.readBoolEntry("ebcomportamiento/parseModulesOnLoad", False)
        for root, dirs, files in os.walk(directorio):
            # print("*", root, dirs, files)
            root_dirs_list = pathlib.Path(root)
            for name in files:
                if name.endswith(extension):
                    if "test" in root_dirs_list.parts or name.startswith("test_"):
                        continue

                    path_ = os.path.join(root, name)

                    if parse_modules_on_load and name.endswith(".qs"):
                        file_py_path_ = "%s.py" % path_
                        if os.path.exists(file_py_path_):
                            os.remove(file_py_path_)

                        postparse.pythonify([path_])
                        if os.path.exists(file_py_path_):
                            value_py = qsa.File(file_py_path_, "UTF-8").read()
                            if not isinstance(value_py, str):
                                raise Exception("value_py must be string not bytes.")

                            self.load_file_to_db("%s.py" % name[-3], value_py, log, directorio)
                    if name.endswith(".jasper"):
                        file_ = open(path_, "rb")
                        value = file_.read()
                        file_.close()
                    else:
                        encode = "UTF-8" if path_.endswith((".ts", ".py")) else "ISO-8859-1"
                        try:
                            value = qsa.File(path_, encode).read()  # type: ignore [assignment]
                        except UnicodeDecodeError:
                            LOGGER.warning(
                                "The file %s has a incorrect encode (%s)" % (path_, encode)
                            )
                            encode = "UTF8" if encode == "ISO-8859-1" else "ISO-8859-1"
                            value = qsa.File(path_, encode).read()  # type: ignore [assignment]

                    self.load_file_to_db(name, value, log, directorio)
                    # qsa.sys.processEvents()

    def load_button_clicked(self) -> None:
        """Load a directory from file system."""
        directorio = qsa.FileDialog.getExistingDirectory(
            "", qsa.util.translate("scripts", "Elegir Directorio")
        )
        self.load_from_disk(directorio or "", True)

    def expoort_button_clicked(self) -> None:
        """Export a module to file system."""
        directorio = qsa.FileDialog.getExistingDirectory(
            "", qsa.util.translate("scripts", "Elegir Directorio")
        )
        self.export_to_disk(directorio or "")

    def accept_license(self, directorio: str) -> bool:
        """Accept license dialog."""
        path_licencia = qsa.Dir.cleanDirPath(qsa.ustr(directorio, "/COPYING"))
        if not qsa.FileStatic.exists(path_licencia):
            qsa.MessageBox.critical(
                qsa.util.translate(
                    "scripts",
                    qsa.ustr(
                        "El fichero ",
                        path_licencia,
                        " con la licencia del módulo no existe.\nEste fichero debe existir para poder aceptar la licencia que contiene.",
                    ),
                ),
                qsa.MessageBox.Ok,
            )
            return False
        licencia = qsa.FileStatic.read(path_licencia)
        dialog = qsa.Dialog()
        dialog.setWidth(600)
        dialog.caption = qsa.util.translate("scripts", "Acuerdo de Licencia.")
        # dialog.newTab(qsa.util.translate(u"scripts", u"Acuerdo de Licencia."))
        texto = qsa.TextEdit()
        texto.text = licencia
        dialog.add(texto)
        dialog.okButtonText = qsa.util.translate("scripts", "Sí, acepto este acuerdo de licencia.")
        dialog.cancelButtonText = qsa.util.translate(
            "scripts", "No, no acepto este acuerdo de licencia."
        )
        return True if dialog.exec() else False

    def load_from_disk(self, directorio: str, check_license: bool) -> None:
        """Load a folder from file system."""
        if directorio:
            if check_license:
                if not self.accept_license(directorio):
                    qsa.MessageBox.critical(
                        qsa.util.translate(
                            "scripts",
                            "Imposible cargar el módulo.\nLicencia del módulo no aceptada.",
                        ),
                        qsa.MessageBox.Ok,
                    )
                    return

            # qsa.sys.cleanupMetaData()
            qsa.sys.processEvents()
            if self.cursor().commitBuffer():
                id_mod_widget = self.child("idMod")
                if id_mod_widget is not None:
                    id_mod_widget.setDisabled(True)
                log = self.child("log")

                if log is None:
                    raise Exception("log is empty!.")

                log.text = ""
                self.setDisabled(True)

                extensiones = [
                    "xml",
                    "mod",
                    "signatures",
                    "certificates",
                    "checksum",
                    "ui",
                    "mtd",
                    "qs",
                    "py",
                    "qry",
                    "kut",
                    "ar",
                    "ts",
                    "jasper",
                    "jrxml",
                ]
                for extension in extensiones:
                    self.load_files(directorio, ".%s" % extension)

                log.append(qsa.util.translate("scripts", "* Carga finalizada."))
                self.setDisabled(False)
                tdb_lineas = self.child("lineas")
                if tdb_lineas is not None:
                    tdb_lineas.refresh()

    def file_type(self, nombre: str) -> str:
        """Return file type."""
        dot_pos = nombre.rfind(".")
        return nombre[dot_pos:]

    def export_to_disk(self, directorio: str) -> None:
        """Export a module to disk."""
        if directorio:
            tdb_lineas = self.child("lineas")
            if tdb_lineas is None:
                raise Exception("lineas control not found")

            cur_files = tdb_lineas.cursor()
            cur_modules = qsa.FLSqlCursor("flmodules")
            cur_areas = qsa.FLSqlCursor("flareas")
            if cur_files.size() != 0:
                dir = qsa.Dir()
                id_modulo = self.cursor().valueBuffer("idmodulo")
                log = self.child("log")
                if log is None:
                    raise Exception("Log control not found!.")

                log.text = ""
                directorio = qsa.Dir.cleanDirPath(qsa.ustr(directorio, "/", id_modulo))
                if not dir.fileExists(directorio):
                    dir.mkdir(directorio)
                if not dir.fileExists(qsa.ustr(directorio, "/forms")):
                    dir.mkdir(qsa.ustr(directorio, "/forms"))
                if not dir.fileExists(qsa.ustr(directorio, "/scripts")):
                    dir.mkdir(qsa.ustr(directorio, "/scripts"))
                if not dir.fileExists(qsa.ustr(directorio, "/queries")):
                    dir.mkdir(qsa.ustr(directorio, "/queries"))
                if not dir.fileExists(qsa.ustr(directorio, "/tables")):
                    dir.mkdir(qsa.ustr(directorio, "/tables"))
                if not dir.fileExists(qsa.ustr(directorio, "/reports")):
                    dir.mkdir(qsa.ustr(directorio, "/reports"))
                if not dir.fileExists(qsa.ustr(directorio, "/translations")):
                    dir.mkdir(qsa.ustr(directorio, "/translations"))
                cur_files.first()
                # file = None
                # tipo = None
                # contenido = ""
                self.setDisabled(True)
                s01_dowhile_1stloop = True
                while s01_dowhile_1stloop or cur_files.next():
                    s01_dowhile_1stloop = False
                    file_name = cur_files.valueBuffer("nombre")
                    tipo = self.file_type(file_name)
                    contenido = (
                        cur_files.valueBuffer("binario")
                        if tipo == ".jasper"
                        else cur_files.valueBuffer("contenido")
                    )

                    if contenido:
                        codec: str = ""
                        if tipo in [
                            ".xml",
                            ".mod",
                            ".xml",
                            ".signatures",
                            ".certificates",
                            ".checksum",
                            ".ui",
                            ".qs",
                            ".qry",
                            ".mtd",
                            ".kut",
                            ".jasper",
                            ".jrxml",
                        ]:
                            codec = "ISO-8859-1"
                        elif tipo in [".py", ".ts"]:
                            codec = "UTF-8"
                        else:
                            log.append(
                                qsa.util.translate(
                                    "scripts", qsa.ustr("* Omitiendo ", file_name, ".")
                                )
                            )
                            continue

                        sub_carpeta = ""
                        if tipo in (".py", ".qs"):
                            sub_carpeta = "scripts"
                        elif tipo == ".qry":
                            sub_carpeta = "queries"
                        elif tipo == ".mtd":
                            sub_carpeta = "tables"
                        elif tipo == ".ts":
                            sub_carpeta = "translations"
                        elif tipo == ".ui":
                            sub_carpeta = "forms"
                        elif tipo in (".kut", ".jasper", ".jrxml"):
                            sub_carpeta = "reports"

                        file_path = os.path.join(directorio, sub_carpeta, file_name)

                        if tipo == ".jasper":
                            file_ = open(file_path, "wb")
                            file_.write(contenido)
                            file_.close()
                        else:
                            qsa.sys.write(codec, file_path, contenido)
                        log.append(
                            qsa.util.translate("scripts", qsa.ustr("* Exportando ", file_name, "."))
                        )

                    # qsa.sys.processEvents()

                cur_modules.select(qsa.ustr("idmodulo = '", id_modulo, "'"))
                if cur_modules.first():
                    cur_areas.select(qsa.ustr("idarea = '", cur_modules.valueBuffer("idarea"), "'"))
                    cur_areas.first()
                    name_area = cur_areas.valueBuffer("descripcion")
                    if not qsa.FileStatic.exists(
                        qsa.ustr(directorio, "/", cur_modules.valueBuffer("idmodulo"), ".xpm")
                    ):
                        qsa.sys.write(
                            "ISO-8859-1",
                            qsa.ustr(directorio, "/", cur_modules.valueBuffer("idmodulo"), ".xpm"),
                            cur_modules.valueBuffer("icono"),
                        )
                        log.append(
                            qsa.util.translate(
                                "scripts",
                                qsa.ustr(
                                    "* Exportando ",
                                    cur_modules.valueBuffer("idmodulo"),
                                    ".xpm (Regenerado).",
                                ),
                            )
                        )
                    if not qsa.FileStatic.exists(
                        qsa.ustr(directorio, "/", cur_modules.valueBuffer("idmodulo"), ".mod")
                    ):
                        contenido = qsa.ustr(
                            "<!DOCTYPE MODULE>\n<MODULE>\n<name>",
                            cur_modules.valueBuffer("idmodulo"),
                            '</name>\n<alias>QT_TRANSLATE_NOOP("FLWidgetApplication","',
                            cur_modules.valueBuffer("descripcion"),
                            '")</alias>\n<area>',
                            cur_modules.valueBuffer("idarea"),
                            '</area>\n<name_area>QT_TRANSLATE_NOOP("FLWidgetApplication","',
                            name_area,
                            '")</name_area>\n<version>',
                            cur_modules.valueBuffer("version"),
                            "</version>\n<icon>",
                            cur_modules.valueBuffer("idmodulo"),
                            ".xpm</icon>\n<flversion>",
                            cur_modules.valueBuffer("version"),
                            "</flversion>\n<description>",
                            cur_modules.valueBuffer("idmodulo"),
                            "</description>\n</MODULE>",
                        )
                        qsa.sys.write(
                            "ISO-8859-1",
                            qsa.ustr(directorio, "/", cur_modules.valueBuffer("idmodulo"), ".mod"),
                            contenido,
                        )
                        log.append(
                            qsa.util.translate(
                                "scripts",
                                qsa.ustr(
                                    "* Generando ",
                                    cur_modules.valueBuffer("idmodulo"),
                                    ".mod (Regenerado).",
                                ),
                            )
                        )

                self.setDisabled(False)
                log.append(qsa.util.translate("scripts", "* Exportación finalizada."))



"""Test_flmanager module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing

from pineboolib import application


class TestFLManagerModules(unittest.TestCase):
    """TestFLManager Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self) -> None:
        """Basic test."""

        manager_modules_ = application.PROJECT.conn_manager.managerModules()
        self.assertEqual(manager_modules_.listAllIdModules(), ["sys"])

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# -*- coding: utf-8 -*-
"""
Static loader emulating Eneboo.

Performs load of scripts from disk instead of database.
"""


from PyQt6 import QtWidgets, QtCore

from pineboolib.core import settings, decorators
from pineboolib.core.utils import logging, utils_base

from pineboolib import application

import os
import importlib
import sys
from typing import List, Optional, cast, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import iconnection  # pragma: no cover

LOGGER = logging.get_logger(__name__)

SHOW_REINIT_MESSAGE = True


class AQStaticDirInfo(object):
    """Store information about a filesystem folder."""

    active_: bool
    path_: str

    def __init__(self, *args) -> None:
        """Inicialize."""

        self.active_ = str(args[0]) == "True"
        self.path_ = "" if len(args) == 1 else args[1]


class AQStaticBdInfo(object):
    """Get or set settings on database related to staticloader."""

    enabled_: bool
    dirs_: List[AQStaticDirInfo]
    key_: str

    def __init__(self, database: "iconnection.IConnection") -> None:
        """Create new AQStaticBdInfo."""
        self.db_ = database.DBName()
        self.dirs_ = []
        self.key_ = "StaticLoader/%s/" % self.db_
        self.enabled_ = settings.CONFIG.value("%senabled" % self.key_, False)

    def findPath(self, path: str) -> Optional[AQStaticDirInfo]:
        """Find if path "path" is managed in this class."""
        for info in self.dirs_:
            if info.path_ == path:
                return info

        return None

    def readSettings(self) -> None:
        """Read settings for staticloader."""
        self.enabled_ = settings.CONFIG.value("%senabled" % self.key_, False)
        self.dirs_.clear()
        dirs = settings.CONFIG.value("%sdirs" % self.key_, [])
        i = 0

        while i < len(dirs):
            active_ = dirs[i]
            i += 1
            path_ = dirs[i]
            i += 1
            self.dirs_.append(AQStaticDirInfo(active_, path_))

    def writeSettings(self) -> None:
        """Write settings for staticloader."""
        settings.CONFIG.set_value("%senabled" % self.key_, self.enabled_)
        dirs: List[Union[bool, str]] = []
        active_dirs = []

        for info in self.dirs_:
            dirs.append(info.active_)
            dirs.append(info.path_)
            if info.active_:
                active_dirs.append(info.path_)

        settings.CONFIG.set_value("%sdirs" % self.key_, dirs)
        settings.CONFIG.set_value("%sactiveDirs" % self.key_, ",".join(active_dirs))

    def msg_static_changed(self, event) -> None:
        """Show reinit msg."""

        global SHOW_REINIT_MESSAGE
        if not isinstance(event.src_path, str) or not event.src_path.find(".") > -1:
            return

        src_upper: str = event.src_path.upper()
        type_upper: str = event.event_type.upper()

        if src_upper.find("__PYCACHE__") > -1:
            return
        if src_upper.find(".MYPY_CACHE") > -1:
            return
        if type_upper in (
            "OPENED",
            "CLOSED_NO_WRITE",
        ):
            return

        if not SHOW_REINIT_MESSAGE:
            return

        SHOW_REINIT_MESSAGE = False

        if utils_base.is_library():
            LOGGER.warning(
                "STATIC LOADER:  %s HAS BEEN %s. REINIT!",
                src_upper,
                type_upper,
            )

            while application.PROJECT.aq_app._inicializing:
                QtWidgets.QApplication.processEvents()

            for key in list(sys.modules.keys()):
                # Si el modulo esta en application.EXTERNAL_FOLDER. lo recargamos
                file_name = (
                    os.path.abspath(sys.modules[key].__file__)
                    if hasattr(sys.modules[key], "__file__")
                    and sys.modules[key].__file__ is not None
                    else None
                )
                if (
                    file_name
                    and application.EXTERNAL_FOLDER
                    and file_name.startswith(application.EXTERNAL_FOLDER)
                ):
                    try:
                        LOGGER.warning(
                            "STATIC LOADER: Reloading external module %s -> %s" % (key, file_name)
                        )
                        importlib.reload(sys.modules[key])
                    except Exception as error:
                        LOGGER.warning(
                            "STATIC LOADER: Error reloading external module %s, Error: %s"
                            % (key, str(error))
                        )

            application.PROJECT.aq_app.reinit()
        else:
            LOGGER.warning(
                "STATIC LOADER:  %s HAS BEEN %s.", event.src_path.upper(), event.event_type.upper()
            )


class FLStaticLoaderWarning(QtCore.QObject):
    """Create warning about static loading."""

    warns_: List[str]
    paths_: List[str]

    def __init__(self) -> None:
        """Create a new FLStaticLoaderWarning."""
        super().__init__()
        self.warns_ = []
        self.paths_ = []

    def popupWarnings(self) -> None:
        """Show a popup if there are any warnings."""
        if not self.warns_:
            return

        msg = '<p><img source="about.png" align="right"><b><u>CARGA ESTATICA ACTIVADA</u></b><br><br><font face="Monospace">'
        for item in self.warns_:
            msg += "%s<br>" % (item)

        msg += "</font><br></p>"
        self.warns_.clear()

        application.PROJECT.aq_app.popupWarn(msg)


WARN_: Optional[FLStaticLoaderWarning] = None


class PNStaticLoader(QtCore.QObject):
    """Perform static loading of scripts from filesystem."""

    def __init__(self, info: "AQStaticBdInfo", dialog: QtWidgets.QDialog) -> None:
        """Create a new FLStaticLoader."""

        super(PNStaticLoader, self).__init__()

        self._dialog = dialog
        self._info = info
        self._dialog.pixOn.setVisible(  # type: ignore[attr-defined] # noqa: F821
            self._info.enabled_
        )
        self._dialog.pixOff.setVisible(  # type: ignore[attr-defined] # noqa: F821
            not self._info.enabled_
        )

        tbl_dir = self._dialog.tblDirs  # type: ignore[attr-defined] # noqa: F821
        tbl_dir.show()
        cast(QtWidgets.QTableWidget, tbl_dir).verticalHeader().setVisible(True)  # type: ignore [union-attr]
        cast(QtWidgets.QTableWidget, tbl_dir).horizontalHeader().setVisible(True)  # type: ignore [union-attr]

        tbl_dir.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        tbl_dir.setAlternatingRowColors(True)
        tbl_dir.setColumnCount(2)
        tbl_dir.setHorizontalHeaderLabels([self.tr("Carpeta"), self.tr("Activo")])

        horizontal_header = tbl_dir.horizontalHeader()
        horizontal_header.setSectionsClickable(False)
        horizontal_header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Fixed)

        horizontal_header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Stretch)
        horizontal_header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)

        self.load()
        cast(
            QtWidgets.QToolButton, self._dialog.pbAddDir  # type: ignore[attr-defined] # noqa: F821
        ).clicked.connect(self.addDir)
        cast(
            QtWidgets.QToolButton, self._dialog.pbModDir  # type: ignore[attr-defined] # noqa: F821
        ).clicked.connect(self.modDir)
        cast(
            QtWidgets.QToolButton, self._dialog.pbDelDir  # type: ignore[attr-defined] # noqa: F821
        ).clicked.connect(self.delDir)
        cast(
            QtWidgets.QToolButton, self._dialog.pbNo  # type: ignore[attr-defined] # noqa: F821
        ).clicked.connect(self._dialog.reject)
        cast(
            QtWidgets.QToolButton, self._dialog.pbOk  # type: ignore[attr-defined] # noqa: F821
        ).clicked.connect(self._dialog.accept)
        cast(
            QtWidgets.QCheckBox, self.chkEnabled
        ).toggled.connect(  # type: ignore [attr-defined] # noqa: F821
            self.setEnabled
        )

    @decorators.pyqt_slot()
    def load(self) -> None:
        """Load and initialize the object."""
        info = self._info
        info.readSettings()
        cast(
            QtWidgets.QLabel, self._dialog.lblBdTop  # type: ignore[attr-defined] # noqa: F821
        ).setText(info.db_)
        cast(
            QtWidgets.QCheckBox, self._dialog.chkEnabled  # type: ignore[attr-defined] # noqa: F821
        ).setChecked(info.enabled_)
        tbl_dir = cast(
            QtWidgets.QTableWidget, self._dialog.tblDirs  # type: ignore[attr-defined] # noqa: F821
        )
        if info.dirs_:
            n_rows = tbl_dir.rowCount()
            if n_rows > 0:
                tbl_dir.clear()

            n_rows = len(info.dirs_)
            tbl_dir.setRowCount(n_rows)

            for row, info_dir in enumerate(info.dirs_):
                item = QtWidgets.QTableWidgetItem(info_dir.path_)
                item.setTextAlignment(
                    QtCore.Qt.AlignmentFlag.AlignVCenter + QtCore.Qt.AlignmentFlag.AlignLeft
                )
                tbl_dir.setItem(row, 0, item)
                chk = QtWidgets.QCheckBox(tbl_dir)
                chk.setChecked(info_dir.active_)
                chk.toggled.connect(self.setChecked)  # type: ignore [attr-defined] # noqa: F821
                tbl_dir.setCellWidget(row, 1, chk)

            tbl_dir.setCurrentCell(n_rows, 0)

    @decorators.pyqt_slot(bool)
    def addDir(self) -> None:
        """Ask user for adding a new folder for static loading."""

        tbl_dir = cast(
            QtWidgets.QTableWidget, self._dialog.tblDirs  # type: ignore[attr-defined] # noqa: F821
        )
        cur_row = tbl_dir.currentRow()
        dir_init = tbl_dir.item(cur_row, 0).text() if cur_row > -1 else ""  # type: ignore [union-attr]

        dir = QtWidgets.QFileDialog.getExistingDirectory(
            None, self.tr("Selecciones el directorio a insertar"), dir_init
        )

        if dir:
            n_rows = tbl_dir.rowCount()
            tbl_dir.setRowCount(n_rows + 1)

            item = QtWidgets.QTableWidgetItem(str(dir))
            item.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignVCenter + QtCore.Qt.AlignmentFlag.AlignLeft
            )
            tbl_dir.setItem(n_rows, 0, item)

            chk = QtWidgets.QCheckBox(tbl_dir)
            chk.setChecked(True)
            chk.toggled.connect(self.setChecked)  # type: ignore [attr-defined] # noqa: F821

            tbl_dir.setCellWidget(n_rows, 1, chk)
            tbl_dir.setCurrentCell(n_rows, 0)

            self._info.dirs_.append(AQStaticDirInfo(True, dir))

    @decorators.pyqt_slot()
    def modDir(self) -> None:
        """Ask user for a folder to change."""

        tbl_dir = cast(QtWidgets.QTableWidget, self.tblDirs)
        cur_row = tbl_dir.currentRow()
        if cur_row == -1:
            return

        actual_dir = tbl_dir.item(cur_row, 0).text() if cur_row > -1 else ""  # type: ignore [union-attr]

        new_dir = QtWidgets.QFileDialog.getExistingDirectory(
            None, self.tr("Selecciones el directorio a modificar"), actual_dir
        )

        if new_dir:
            info = self._info.findPath(actual_dir)
            if info:
                info.path_ = new_dir

            item = QtWidgets.QTableWidgetItem(str(new_dir))
            item.setTextAlignment(
                QtCore.Qt.AlignmentFlag.AlignVCenter + QtCore.Qt.AlignmentFlag.AlignLeft
            )
            tbl_dir.setItem(cur_row, 0, item)

    @decorators.pyqt_slot()
    def delDir(self) -> None:
        """Ask user for folder to delete."""

        tbl_dir = cast(QtWidgets.QTableWidget, self.tblDirs)
        cur_row = tbl_dir.currentRow()
        if cur_row == -1:
            return

        if QtWidgets.QMessageBox.StandardButton.No == QtWidgets.QMessageBox.warning(
            QtWidgets.QWidget(),
            self.tr("Borrar registro"),
            self.tr("El registro activo será borrado. ¿ Está seguro ?"),
            QtWidgets.QMessageBox.StandardButton.Ok,
            QtWidgets.QMessageBox.StandardButton.No,
        ):
            return

        info = self._info.findPath(tbl_dir.item(cur_row, 0).text())  # type: ignore [union-attr]
        if info:
            self._info.dirs_.remove(info)

        tbl_dir.removeRow(cur_row)

    @decorators.pyqt_slot(bool)
    def setEnabled(self, state: bool) -> None:
        """Enable or disable this object."""
        self._info.enabled_ = state
        self._dialog.pixOn.setVisible(state)  # type: ignore[attr-defined] # noqa: F821
        self._dialog.pixOff.setVisible(not state)  # type: ignore[attr-defined] # noqa: F821

    @decorators.pyqt_slot(bool)
    def setChecked(self, state: bool) -> None:
        """Set checked this object."""

        tbl_dir = cast(QtWidgets.QTableWidget, self.tblDirs)
        chk = self.sender()
        if not chk:
            return

        for row in range(tbl_dir.rowCount()):
            if tbl_dir.cellWidget(row, 1) is chk:
                info = self._info.findPath(tbl_dir.item(row, 0).text())  # type: ignore [union-attr]
                if info:
                    info.active_ = state

    @staticmethod
    def setup(info: "AQStaticBdInfo", dialog: QtWidgets.QDialog) -> None:
        """Configure user interface from given widget."""

        diag_setup = PNStaticLoader(info, dialog)
        if QtWidgets.QDialog.DialogCode.Accepted == diag_setup._dialog.exec():
            info.writeSettings()

    @staticmethod
    def content(name: str, info: "AQStaticBdInfo", only_path: bool = False) -> str:
        """Get content from given path."""
        global WARN_
        info.readSettings()
        candidates: List[List[str]] = []

        if not info.dirs_:
            LOGGER.info(
                "STATIC LOAD: No folders found searching %s. Please disable static load or add folders",
                name,
            )
            return ""

        for info_item in info.dirs_:
            content_path_candidate = os.path.join(info_item.path_, name)
            if info_item.path_ not in content_path_candidate:
                continue
            if info_item.active_ and os.path.exists(
                content_path_candidate
            ):  # Buscamos todos los candidatos
                if content_path_candidate in [
                    path_candidate for name, path_candidate in candidates
                ]:
                    continue
                candidates.append([info_item.path_, content_path_candidate])

        if candidates:
            if not WARN_:
                WARN_ = FLStaticLoaderWarning()

            file_path, content_path = candidates[0]
            msg = "%s -> ...%s" % (name, file_path[0:40])

            if msg not in WARN_.warns_:
                if len(candidates) > 1:  # Si hay mas de un candidato muestra warning
                    LOGGER.warning(
                        "STATIC LOAD: MULTIPLES CANIDATES FOUND FOR %s. USING FIRST:", name.upper()
                    )
                    for num, candidate in enumerate(candidates):
                        LOGGER.warning("    %s) - %s/%s", num + 1, candidate[0], name)

                WARN_.warns_.append(msg)
                WARN_.paths_.append("%s:%s" % (name, file_path))
                if settings.CONFIG.value("ebcomportamiento/SLConsola", False):
                    LOGGER.warning("STATIC LOAD: ACTIVATED %s -> %s", name, file_path)
                if settings.CONFIG.value("ebcomportamiento/SLGUI", False):
                    QtCore.QTimer.singleShot(500, WARN_.popupWarnings)

            if only_path:
                return content_path
            else:
                return application.PROJECT.conn_manager.managerModules().contentFS(
                    os.path.join(file_path, name)
                )

        return ""

    def __getattr__(self, name: str) -> QtWidgets.QWidget:
        """Emulate child properties as if they were inserted into the object."""
        return cast(QtWidgets.QWidget, self._dialog.findChild(QtWidgets.QWidget, name))



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""FLmetadata_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flmetadata(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flmetadata class."""

    __tablename__ = "flmetadata"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flmetadata",
        "alias": "Metadatos",
        "fields": [
            {
                "name": "tabla",
                "alias": "Nombre de la tabla",
                "pk": True,
                "type": "string",
                "length": 255,
                "null": False,
            },
            {"name": "xml", "alias": "Descripción XML", "type": "stringlist", "visiblegrid": False},
            {"name": "bloqueo", "alias": "Tabla bloqueada", "type": "bool"},
            {"name": "seq", "alias": "Secuencia", "type": "uint", "null": False},
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    tabla = sqlalchemy.Column("tabla", sqlalchemy.String(255), primary_key=True)
    xml = sqlalchemy.Column("xml", sqlalchemy.String)
    bloqueo = sqlalchemy.Column("bloqueo", sqlalchemy.Boolean)
    seq = sqlalchemy.Column("seq", sqlalchemy.BigInteger)


# <--- Fields ---



"""Fldatatable module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtCore, QtWidgets, QtGui  # type: ignore[import]

from pineboolib.core import decorators, settings
from pineboolib.core.utils import utils_base

from pineboolib import logging

from typing import Any, Optional, List, Dict, Tuple, cast, TYPE_CHECKING

from pineboolib.application.database import pnsqlcursor, pncursortablemodel


if TYPE_CHECKING:
    from pineboolib.application.metadata import pnfieldmetadata  # pragma: no cover
    from pineboolib.application.metadata import pntablemetadata  # pragma: no cover
    from pineboolib.interfaces import isqlcursor  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class FLDataTable(QtWidgets.QTableView):
    """
    Class that is a redefinition of the QDataTable class.

    Specifies for the needs of AbanQ.
    """

    _parent: Optional[Any]
    filter_: str
    sort_: str
    fltable_iface: Any

    """
    Numero de la fila (registro) seleccionada actualmente
    """
    row_selected: int

    """
    Numero de la columna (campo) seleccionada actualmente
    """
    col_selected: int

    """
    Cursor, con los registros
    """
    cursor_: Optional["isqlcursor.ISqlCursor"]

    """
    Almacena la tabla está en modo sólo lectura
    """
    readonly_: bool

    """
    Almacena la tabla está en modo sólo edición
    """
    editonly_: bool

    """
    Indica si la tabla está en modo sólo inserción
    """
    insertonly_: bool

    """
    Texto del último campo dibujado en la tabla
    """
    lastTextPainted_: Optional[str]

    """
    Pixmap precargados
    """
    pix_ok_: str
    pix_no_: str

    """
    Lista con las claves primarias de los registros seleccionados por chequeo
    """
    pk_checked: List[object]

    """
    Filtro persistente para el cursor
    """
    persistent_filter_: str

    """
    Indicador para evitar refrescos anidados
    """
    refreshing_: bool

    """
    Indica si el componente es emergente ( su padre es un widget del tipo Popup )
    """
    popup_: bool

    """
    Indica el ancho de las columnas establecidas explícitamente con FLDataTable::setColumnWidth
    """
    width_cols_: Dict[str, int]

    """
    Indica si se deben mostrar los campos tipo pixmap en todas las filas
    """
    show_all_pixmaps_: bool

    """
    Nombre de la función de script a invocar para obtener el color de las filas y celdas
    """
    function_get_color: Optional[str]

    """
    Indica que no se realicen operaciones con la base de datos (abrir formularios). Modo "sólo tabla".
    """
    only_table_: bool
    changing_num_rows_: bool
    _paint_field_name: Optional[str]
    paint_field_mtd_: Optional["pnfieldmetadata.PNFieldMetaData"]

    def __init__(
        self, parent: Optional[Any] = None, name: str = "FLDataTable", popup: bool = False
    ):
        """Inicialize."""

        super().__init__(parent)

        if parent:
            self._parent = parent

        self.setObjectName(name)

        self.readonly_ = False
        self.editonly_ = False
        self.insertonly_ = False
        # self.refreshing_ = False
        self.filter_ = ""
        self.sort_ = ""
        self.row_selected = -1
        self.col_selected = -1
        self.pk_checked = []
        self.persistent_filter_ = ""
        self.width_cols_ = {}

        self.pix_ok_ = utils_base.filedir("./core/images/icons", "unlock.png")
        self.pix_no_ = utils_base.filedir("./core/images/icons", "lock.png")
        self.paint_field_mtd_ = None
        self.refreshing_ = False
        self.popup_ = False
        self.show_all_pixmaps_ = False
        self.only_table_ = False
        self.function_get_color = None
        self.changing_num_rows_ = False
        self.cursor_ = None

        self._v_header = self.verticalHeader()
        if self._v_header:
            self._v_header.setDefaultSectionSize(22)

        self._h_header = self.horizontalHeader()
        if self._h_header:
            self._h_header.setDefaultSectionSize(120)
            self._h_header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeMode.Interactive)
        self.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.SingleSelection)
        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(True)
        self.sortByColumn(0, QtCore.Qt.SortOrder.AscendingOrder)
        self.fltable_iface = None
        self.popup_ = popup

    # def __del__(self) -> None:
    #    """Destroyer."""

    # if self.timerViewRepaint_:
    #    self.timerViewRepaint_.stop()

    #    if self.cursor_:
    #        self.cursor_.restoreEditionFlag(self.objectName())
    #        self.cursor_.restoreBrowseFlag(self.objectName())

    def header(self) -> Any:
        """Return the FLDatatable header."""

        return self._h_header

    def model(self) -> pncursortablemodel.PNCursorTableModel:
        """Return cursor table model."""
        return cast(pncursortablemodel.PNCursorTableModel, super().model())

    def setFLSqlCursor(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Set the cursor."""

        if cursor and cursor.metadata():
            cur_chg = False
            if self.cursor_ and not self.cursor_ == cursor:
                self.cursor_.restoreEditionFlag(self.objectName())
                self.cursor_.restoreBrowseFlag(self.objectName())
                cast(
                    QtCore.pyqtSignal, self.cursor_.commited
                ).disconnect(  # type: ignore [attr-defined] # noqa: F821
                    self.refresh
                )

                cur_chg = True

            if not self.cursor_ or cur_chg:
                self.cursor_ = cursor
                if not self.cursor_:
                    raise Exception("cursor_ is empty!")

                self.cursor_._is_delegate_commit = settings.CONFIG.value(
                    "application/delegateCommit", False
                )

                self.setFLReadOnly(self.readonly_)
                self.setEditOnly(self.editonly_)
                self.setInsertOnly(self.insertonly_)
                self.setOnlyTable(self.only_table_)

                cast(
                    QtCore.pyqtSignal, self.cursor_.commited
                ).connect(  # type: ignore [attr-defined] # noqa: F821
                    self.refresh
                )

                self.setModel(self.cursor_.model())
                selection_model = self.cursor_.selection()
                if selection_model is not None:
                    self.setSelectionModel(selection_model)
                else:
                    LOGGER.warning("Invalid selection model in %s" % self.cursor_.curName())
                # self.model().sort(self.header().logicalIndex(0), 0)
                self.installEventFilter(self)
                self.model().set_parent_view(self)

    def setPersistentFilter(self, p_filter: Optional[str] = None) -> None:
        """Set the persistent filter for this control."""

        if p_filter is None:
            raise Exception("Invalid use of setPersistentFilter with None")
        self.persistent_filter_ = p_filter

    def setFilter(self, filter: str) -> None:
        """Set the filter for this control."""
        self.filter_ = filter

    def numCols(self) -> int:
        """
        Return the number of columns.
        """

        return self.horizontalHeader().count()  # type: ignore [union-attr]

    def setSort(self, sort: str) -> None:
        """Return the ascending / descending order of the first columns."""

        self.sort_ = sort

    # def cursor(self) -> Optional["isqlcursor.ISqlCursor"]:
    #    """
    #    Devuelve el cursor
    #    """
    #    return self.cursor_

    @property
    def cur(self) -> "isqlcursor.ISqlCursor":
        """
        Return the cursor used by the control.
        """

        if self.cursor_ is None:
            raise Exception("Cursor not set yet")
        if self.cursor_.aqWasDeleted():
            raise Exception("Cursor was deleted")
        return self.cursor_

    def setFLReadOnly(self, mode: bool) -> None:
        """
        Set the table to read only or not.
        """

        if not self.cursor_ or self.cursor_.aqWasDeleted():
            return

        self.cursor_.setEdition(not mode, self.objectName())
        self.readonly_ = mode

    def flReadOnly(self) -> bool:
        """
        Return if the table is in read-only mode.
        """

        return self.readonly_

    def setEditOnly(self, mode: bool) -> None:
        """
        Set the table to edit only or not.
        """

        if not self.cursor_ or self.cursor_.aqWasDeleted():
            return

        self.editonly_ = mode

    def editOnly(self) -> bool:
        """
        Return if the table is in edit-only mode.
        """

        return self.editonly_

    def setInsertOnly(self, mode: bool) -> None:
        """
        Set the table to insert only or not.
        """

        if not self.cursor_ or self.cursor_.aqWasDeleted():
            return

        self.cursor_.setEdition(not mode, self.objectName())
        self.insertonly_ = mode

    def insertOnly(self) -> bool:
        """
        Return if the table is in insert-only mode.
        """
        return self.insertonly_

    def primarysKeysChecked(self) -> list:
        """
        Get the list with the primary keys of the records selected by check.
        """

        return self.pk_checked

    def clearChecked(self) -> None:
        """
        Clear the list with the primary keys of the records selected by check.
        """
        self.pk_checked.clear()
        model = self.cur.model()
        for idx in model._check_column.keys():
            model._check_column[idx].setChecked(False)

    def setPrimaryKeyChecked(self, pk_value: str, on_: bool) -> None:
        """
        Set the status selected by check for a record, indicating the value of its primary key.
        """

        model = self.cur.model()
        if on_:
            if pk_value not in self.pk_checked:
                self.pk_checked.append(pk_value)
                self.primaryKeyToggled.emit(pk_value, False)
        else:
            if pk_value in self.pk_checked:
                self.pk_checked.remove(pk_value)
                self.primaryKeyToggled.emit(pk_value, False)

        if pk_value not in model._check_column.keys():
            model._check_column[pk_value] = QtWidgets.QCheckBox()

        model._check_column[pk_value].setChecked(on_)

    def setShowAllPixmaps(self, value: bool) -> None:
        """
        Set if the pixmaps of unselected lines are displayed.
        """

        self.show_all_pixmaps_ = value

    def showAllPixmap(self) -> bool:
        """
        Return if pixmaps of unselected lines are displayed.
        """

        return self.show_all_pixmaps_

    def setFunctionGetColor(self, func_name: Optional[str], iface: Optional[Any] = None) -> None:
        """
        Set the function to use to calculate the color of the cell.
        """

        self.fltable_iface = iface
        self.function_get_color = func_name

    def functionGetColor(self) -> Tuple[Optional[str], Any]:
        """
        Return the function to use to calculate the color of the cell.
        """

        return (self.function_get_color, self.fltable_iface)

    def setOnlyTable(self, on_: bool = True) -> None:
        """
        Set if the control is only Table mode.
        """
        if not self.cursor_ or self.cursor_.aqWasDeleted():
            return

        self.cursor_.setEdition(not on_, self.objectName())
        self.cursor_.setBrowse(not on_, self.objectName())
        self.only_table_ = on_

    def onlyTable(self) -> bool:
        """
        Return if the control is only Table mode.
        """

        return self.only_table_

    def indexOf(self, idx: int) -> int:
        """
        Return the visual index of a position.
        """

        return self.header().visualIndex(idx)

    def fieldName(self, col: int) -> str:
        """
        Return the name of the field according to a position.
        """

        field = self.cur.metadata().indexFieldObject(self.indexOf(col))
        if field is None:
            raise Exception("Field not found")
        return field.name()

    def eventFilter(self, obj: Any, event: Optional["QtCore.QEvent"]) -> bool:
        """
        Event Filtering.
        """

        row = self.currentRow()
        col = self.currentColumn()
        num_rows = self.numRows()
        num_cols = self.numCols()
        if event.type() == QtCore.QEvent.Type.KeyPress:  # type: ignore [union-attr]
            key_event = cast(QtGui.QKeyEvent, event)

            if (
                key_event.key() == cast(int, QtCore.Qt.Key.Key_Escape)
                and self.popup_
                and self.parentWidget()
            ):
                self.parentWidget().hide()  # type: ignore [union-attr]
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Insert):
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_F2):
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Up) and row == 0:
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Left) and col == 0:
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Down) and row == num_rows - 1:
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Right) and col == num_cols - 1:
                return True

            if (
                key_event.key()
                in (cast(int, QtCore.Qt.Key.Key_Enter), cast(int, QtCore.Qt.Key.Key_Return))
                and row > -1
            ):
                self.recordChoosed.emit()
                return True

            if key_event.key() == cast(int, QtCore.Qt.Key.Key_Space):
                self.setChecked(self.model().index(row, col))

            if not settings.CONFIG.value("ebcomportamiento/FLTableShortCut", False):
                if key_event.key() == cast(int, QtCore.Qt.Key.Key_A) and not self.popup_:
                    if (
                        self.cursor_
                        and not self.readonly_
                        and not self.editonly_
                        and not self.only_table_
                    ):
                        self.cursor_.insertRecord()
                        return True
                    else:
                        return False

                if key_event.key() == cast(int, QtCore.Qt.Key.Key_C) and not self.popup_:
                    if (
                        self.cursor_
                        and not self.readonly_
                        and not self.editonly_
                        and not self.only_table_
                    ):
                        self.cursor_.copyRecord()
                        return True
                    else:
                        return False

                if key_event.key() == cast(int, QtCore.Qt.Key.Key_M) and not self.popup_:
                    if self.cursor_ and not self.readonly_ and not self.only_table_:
                        self.cursor_.editRecord()
                        return True
                    else:
                        return False

                if key_event.key() == cast(int, QtCore.Qt.Key.Key_Delete) and not self.popup_:
                    if (
                        self.cursor_
                        and not self.readonly_
                        and not self.editonly_
                        and not self.only_table_
                    ):
                        self.cursor_.deleteRecord()
                        return True
                    else:
                        return False

                if key_event.key() == cast(int, QtCore.Qt.Key.Key_V) and not self.popup_:
                    if self.cursor_ and not self.only_table_:
                        self.cursor_.browseRecord()
                        return True

            return False

        return super().eventFilter(obj, event)  # type: ignore [arg-type]

    def contextMenuEvent(self, event: Any) -> None:
        """
        To prevent the context menu from appearing with the options to edit records.
        """

        super().contextMenuEvent(event)

        if not self.cursor_ or not self.cursor_.isValid() or not self.cursor_.metadata():
            return

        mtd = self.cursor_.metadata()
        pri_key = mtd.primaryKey()

        field = mtd.field(pri_key)
        if field is None:
            return

        rel_list = field.relationList()
        if not rel_list:
            return

        conn_db = self.cursor_.db()
        pri_key_val = self.cursor_.valueBuffer(pri_key)

        from pineboolib.q3widgets.qmenu import QMenu
        from pineboolib.q3widgets.qwidget import QWidget
        from pineboolib.q3widgets.qvboxlayout import QVBoxLayout

        from pineboolib.fllegacy.fldatatable import FLDataTable

        popup = QMenu(self)

        menu_frame = QWidget(self, QtCore.Qt.WindowType.Popup)

        lay = QVBoxLayout()
        menu_frame.setLayout(lay)

        tmp_pos = event.globalPos()

        for rel in rel_list:
            cur = pnsqlcursor.PNSqlCursor(
                rel.foreignTable(), True, conn_db.connectionName(), None, None, popup
            )

            if cur.private_cursor.metadata_:
                mtd = cur.metadata()
                field = mtd.field(rel.foreignField())
                if field is None:
                    continue

                sub_popup = QMenu(self)
                sub_popup.setTitle(mtd.alias())
                sub_popup_frame = QWidget(sub_popup, QtCore.Qt.WindowType.Popup)
                lay_popup = QVBoxLayout(sub_popup)
                sub_popup_frame.setLayout(lay_popup)

                data_table = FLDataTable(None, "FLDataTable", True)
                lay_popup.addWidget(data_table)

                data_table.setFLSqlCursor(cur)
                filter = (
                    conn_db.connManager().manager().formatAssignValue(field, pri_key_val, False)
                )
                cur.setFilter(filter)
                data_table.setFilter(filter)
                data_table.refresh()

                # horiz_header = dt.header()
                for i in range(data_table.numCols()):
                    field = mtd.indexFieldObject(i)
                    if not field:
                        continue

                    if not field.visibleGrid():
                        data_table.setColumnHidden(i, True)

                sub_menu = popup.addMenu(sub_popup)
                sub_menu.hovered.connect(  # type: ignore [attr-defined, arg-type, union-attr] # noqa: F821
                    sub_popup_frame.show
                )
                sub_popup_frame.move(
                    tmp_pos.x() + 200, tmp_pos.y()
                )  # FIXME: Hay que mejorar esto ...

        popup.move(tmp_pos.x(), tmp_pos.y())

        popup.exec(event.globalPos())
        del popup
        event.accept()

    def setChecked(self, index: Any) -> None:
        """
        Behavior when clicking on a cell.
        """

        row = index.row()
        col = index.column()
        field = self.cur.metadata().indexFieldObject(col)
        _type = field.type()
        if _type != "check":
            return
        model = self.cur.model()
        primary_key = model.value(row, self.cur.metadata().primaryKey())
        model._check_column[primary_key].setChecked(
            not model._check_column[primary_key].isChecked()
        )
        self.setPrimaryKeyChecked(str(primary_key), model._check_column[primary_key].isChecked())
        # print("FIXME: falta un repaint para ver el color!!")

    def paintFieldMtd(
        self, field_name: str, table_metadata: "pntablemetadata.PNTableMetaData"
    ) -> "pnfieldmetadata.PNFieldMetaData":
        """
        Return the metadata of a field.
        """

        if self.paint_field_mtd_ and self._paint_field_name == field_name:
            return self.paint_field_mtd_

        self._paint_field_name = field_name
        self.paint_field_mtd_ = table_metadata.field(field_name)

        if self.paint_field_mtd_ is None:
            raise Exception("paint_field_mtd_ is empty!.")

        return self.paint_field_mtd_

    timerViewRepaint_ = None

    # def focusInEvent(self, e: QtGui.QFocusEvent) -> None:
    #    """
    #    Focus pickup event.
    #    """

    #    obj = self
    #    # refresh = True
    #    while obj.parent():
    #        if getattr(obj, "inExec_", False):
    #            # refresh = False
    #            break
    #        else:
    #            obj = obj.parent()

    #    # if refresh:
    #    #    self.refresh()
    #    super().focusInEvent(e)

    def sortByColumn(self, column, order) -> None:
        """Overload Sort by column."""

        fix_column = 0

        if self.cursor_ is not None:
            fix_column = self.visual_index_to_metadata_index(column)
            mtdfield = self.model().metadata().indexFieldObject(fix_column)
            if mtdfield is not None:
                self.sort_ = "%s %s" % (mtdfield.name(), "ASC" if not order else "DESC")

        super().sortByColumn(fix_column, order)

    def refresh(self, refresh_option: Any = None) -> None:
        """
        Refresh the cursor.
        """
        if not self.cursor_:
            return

        if self.popup_:
            self.cursor_.refresh()
        # if not self.refreshing_ and self.cursor_ and not self.cursor_.aqWasDeleted() and self.cursor_.metadata():
        if not self.refreshing_:
            # if self.function_get_color and self.cursor_.model:
            #    if self.cursor_.model.color_function_ != self.function_get_color:
            #        self.cursor_.model.setColorFunction(self.function_get_color)

            self.refreshing_ = True
            self.hide()
            filter_: str = self.persistent_filter_
            if self.filter_:
                if self.filter_ not in self.persistent_filter_:
                    if self.persistent_filter_:
                        filter_ = "%s AND %s" % (filter_, self.filter_)
                    else:
                        filter_ = self.filter_

            self.cursor_.setFilter(filter_)

            if self.sort_:
                self.cursor_.setSort(self.sort_)

            last_pk = None
            buffer = self.cursor_.private_cursor.buffer_
            if buffer:
                last_pk = buffer.value(self.cursor_.primaryKey())

            self.cursor_.refresh()
            if last_pk is not None:
                self.selectRow(self.cur.at())
                self.cursor_.refreshBuffer()
            self.show()
            self.refreshing_ = False

    # @decorators.pyqtSlot()
    # @decorators.pyqtSlot(int)
    @decorators.not_implemented_warn
    def ensureRowSelectedVisible(self, position: Optional[int] = None) -> None:
        """Ensure row selected visible."""

        pass

    #    """
    #    Make the selected row visible.
    #    """
    #    print("****", position, self.cursor_.at(), self.cursor_.isValid())

    #    if position is None:
    #        if self.cursor():
    #            position = self.cursor_.at()
    #        else:
    #            return

    # index = self.cursor_.model.index(position, 0)
    # if index is not None:
    #    self.scrollTo(index)

    def setQuickFocus(self) -> None:
        """
        Fast focus without refreshments to optimize.
        """

        # setPaletteBackgroundColor(qApp->palette().color(QPalette::Active, QColorGroup::Base)); FIXME
        super(FLDataTable, self).setFocus()

    def setColWidth(self, field: str, width: int) -> None:
        """
        Set the width of a column.

        @param field Name of the database field corresponding to the column.
        @param width Column width.
        """

        self.width_cols_[field] = width

    def resize_column(self, col: int, str_text: Optional[str]) -> None:
        """
        Resize a column.
        """

        if str_text is None:
            return

        str_text = str(str_text)

        field = self.model().metadata().indexFieldObject(col)
        if field.name() in self.width_cols_.keys():
            if self.columnWidth(col) < self.width_cols_[field.name()]:
                self.header().resizeSection(col, self.width_cols_[field.name()])
        else:
            wc_ = self.header().sectionSize(col)

            font_metrics = QtGui.QFontMetrics(self.header().font())
            wh_ = font_metrics.horizontalAdvance(field.alias() + "W")
            if wh_ < wc_:
                wh_ = wc_

            wc_ = font_metrics.horizontalAdvance(str_text) + font_metrics.maxWidth()
            if wc_ > wh_:
                self.header().resizeSection(col, wc_)
                if col == 0 and self.popup_:
                    parent_widget = self.parentWidget()
                    if parent_widget and parent_widget.width() < wc_:
                        self.resize(wc_, parent_widget.height())
                        parent_widget.resize(wc_, parent_widget.height())

    def cursorDestroyed(self, obj: Optional[Any] = None) -> None:
        """
        Unlink a cursor to this control.
        """

        if not obj or not isinstance(obj, pnsqlcursor.PNSqlCursor):
            return

        self.cursor_ = None

    """
    Indicate that a record has been chosen
    """
    recordChoosed = QtCore.pyqtSignal()
    """
    Indicate that the status of the record selection field has changed.

    That is to say your primary key has been included or removed from the list of selected primary keys.
    This signal is emitted when the user clicks on the check control and when it is changed
    Programmatically check using the FLDataTable :: setPrimaryKeyChecked method.

    @param primaryKeyValue The value of the primary key of the corresponding record.
    @param on The new state; TRUE check activated, FALSE check disabled.
    """
    primaryKeyToggled = QtCore.pyqtSignal(str, bool)

    def numRows(self) -> int:
        """
        Return number of records offered by the cursor.
        """

        if not self.cursor_:
            return -1

        return self.cursor_.model().rowCount()

    def column_name_to_column_index(self, name: str) -> int:
        """
        Return the real index (incusive hidden columns) from a field name.

        @param name The name of the field to look for in the table.
        @return column position in the table.
        """

        if not self.cursor_:
            return -1

        return self.cursor_.model().metadata().fieldIsIndex(name)

    def mouseDoubleClickEvent(self, event: Optional["QtGui.QMouseEvent"]) -> None:
        """Double click event."""
        if cast(QtGui.QMouseEvent, event).button() != QtCore.Qt.MouseButton.LeftButton:
            return

        self.recordChoosed.emit()

    def visual_index_to_column_index(self, col_pos: int) -> int:
        """
        Return the column index from an index of visible columns.

        @param col_pos visible column position.
        @return index column of the column.
        """

        if not self.cursor_:
            return -2

        visible_id = -1
        ret_ = -1
        for column in range(self.model().columnCount()):
            if not self.isColumnHidden(self.logical_index_to_visual_index(column)):
                visible_id += 1

                if visible_id == col_pos:
                    ret_ = column
                    break

        return ret_

    def visual_index_to_metadata_index(self, col_pos: int) -> int:
        """
        Visual to logical index.
        """
        return self.header().logicalIndex(col_pos)

    def logical_index_to_visual_index(self, col_pos: int) -> int:
        """
        Logical Index to Visual Index.
        """
        return self.header().visualIndex(col_pos)

    def visual_index_to_field(self, pos_: int) -> Optional["pnfieldmetadata.PNFieldMetaData"]:
        """
        Return the metadata of a field according to visual position.
        """
        valid_idx = self.visual_index_to_column_index(pos_)  # corrige posición con ocultos.
        logical_idx = self.visual_index_to_metadata_index(valid_idx)  # posicion en metadata.
        if logical_idx == -1:
            return None

        table_metadata = self.model().metadata()
        field_metadata = table_metadata.indexFieldObject(logical_idx)

        if not field_metadata.visibleGrid():
            raise ValueError(
                "Se ha devuelto el field %s.%s que no es visible en el grid"
                % (table_metadata.name(), field_metadata.name())
            )
        # print("POS", pos_, field_metadata.name())
        return field_metadata

    def currentRow(self) -> int:
        """
        Return the current row.
        """
        return self.currentIndex().row()

    def currentColumn(self) -> int:
        """
        Return the current column.
        """
        return self.currentIndex().column()
